import typing
import decimal

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcLocale"


class _interface(ABC):
    import locale as local_locale

    from builtins import str as _builtins_str

    @property
    @abstractmethod
    def LC_CTYPE(self):
        return self.local_locale.LC_CTYPE

    @property
    @abstractmethod
    def LC_COLLATE(self):
        return self.local_locale.LC_COLLATE

    @property
    @abstractmethod
    def LC_TIME(self):
        return self.local_locale.LC_TIME

    @property
    @abstractmethod
    def LC_MONETARY(self):
        return self.local_locale.LC_MONETARY

    @property
    @abstractmethod
    def LC_NUMERIC(self):
        return self.local_locale.LC_NUMERIC

    @property
    @abstractmethod
    def LC_ALL(self):
        return self.local_locale.LC_ALL

    @property
    @abstractmethod
    def CHAR_MAX(self):
        return self.local_locale.CHAR_MAX

    @property
    @abstractmethod
    def Error(self) -> local_locale.Error:
        ...

    @property
    @abstractmethod
    def locale_alias(self) -> dict[_builtins_str, _builtins_str]:
        ...

    @abstractmethod
    def getlocale(self, category: int = ...) -> tuple[_builtins_str | None, _builtins_str | None]:
        ...

    @abstractmethod
    def getdefaultlocale(self, envvars: tuple[_builtins_str, ...] = ...) -> tuple[_builtins_str | None, _builtins_str | None]:
        ...

    @abstractmethod
    def getpreferredencoding(self, do_setlocale: bool = ...) -> _builtins_str:
        ...

    @abstractmethod
    def setlocale(self, do_setlocale: bool = ...) -> _builtins_str:
        ...

    @abstractmethod
    def resetlocale(self, category: int = ...) -> None:
        ...

    @abstractmethod
    def localeconv(
        self,
    ) -> typing.Mapping[_builtins_str, int | _builtins_str | list[int]]:
        ...

    @abstractmethod
    def strcoll(self, __os1: _builtins_str, __os2: _builtins_str, /) -> int:
        ...

    @abstractmethod
    def strxfrm(self, __string: _builtins_str, /) -> _builtins_str:
        ...

    @abstractmethod
    def str(self, val: float) -> _builtins_str:
        ...

    @abstractmethod
    def atof(self, string: _builtins_str, func: typing.Callable[[_builtins_str], float]) -> float:
        ...

    @abstractmethod
    def atoi(self, string: _builtins_str) -> int:
        ...

    @abstractmethod
    def format(
        self,
        percent: _builtins_str,
        value: float | decimal.Decimal,
        grouping: bool = ...,
        monetary: bool = ...,
        *additional: typing.Any,
    ):
        ...

    @abstractmethod
    def format_string(
        self,
        f: _builtins_str,
        val: typing.Any,
        grouping: bool = ...,
        monetary: bool = ...,
    ) -> _builtins_str:
        ...

    @abstractmethod
    def currency(
        self,
        val: int | float | decimal.Decimal,
        symbol: bool = ...,
        grouping: bool = ...,
        international: bool = ...,
    ) -> _builtins_str:
        ...

    @abstractmethod
    def normalize(self, localename: _builtins_str) -> _builtins_str:
        ...

    @abstractmethod
    def _localize(self):
        ...

    @abstractmethod
    def getencoding(self) -> _builtins_str:
        ...

    @abstractmethod
    def delocalize(self, string: _builtins_str) -> _builtins_str:
        ...

    @abstractmethod
    def _parse_localename(self):
        ...


class AbcLocale(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mlocale"

    __all__ = [
        "getlocale",
        "getdefaultlocale",
        "getpreferredencoding",
        "Error",
        "setlocale",
        "resetlocale",
        "localeconv",
        "strcoll",
        "strxfrm",
        "_builtins_str",
        "atof",
        "atoi",
        "format",
        "format_string",
        "currency",
        "normalize",
        "LC_CTYPE",
        "LC_COLLATE",
        "LC_TIME",
        "LC_MONETARY",
        "LC_NUMERIC",
        "LC_ALL",
        "CHAR_MAX",
    ]

    __doc__ = _interface.local_locale.__doc__

    __delegated__ = {_interface: _interface.local_locale}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_locale, self)

        self.__delegate__(_interface, mod)



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcIo"


class io:
    import io

    RawIOBase = io.RawIOBase
    IOBase = io.IOBase
    BufferedIOBase = io.BufferedIOBase
    TextIOBase = io.TextIOBase

    __all__ = [
        "BlockingIOError",
        "open",
        "open_code",
        "IOBase",
        "RawIOBase",
        "FileIO",
        "BytesIO",
        "StringIO",
        "BufferedIOBase",
        "BufferedReader",
        "BufferedWriter",
        "BufferedRWPair",
        "BufferedRandom",
        "TextIOBase",
        "TextIOWrapper",
        "UnsupportedOperation",
        "SEEK_SET",
        "SEEK_CUR",
        "SEEK_END",
    ]

    SEEK_SET = io.SEEK_SET
    SEEK_CUR = io.SEEK_CUR
    SEEK_END = io.SEEK_END

    UnsupportedOperation = io.UnsupportedOperation


class msvcrt:
    setmode: None = None


class _interface(ABC):
    import _pyio as local_io

    @property
    @abstractmethod
    def IOBase(self) -> local_io.IOBase:
        ...

    @property
    @abstractmethod
    def RawIOBase(self) -> local_io.RawIOBase:
        ...

    @property
    @abstractmethod
    def FileIO(self) -> local_io.FileIO:
        ...

    @property
    @abstractmethod
    def BytesIO(self) -> local_io.BytesIO:
        ...

    @property
    @abstractmethod
    def StringIO(self) -> local_io.StringIO:
        ...

    @property
    @abstractmethod
    def BufferedIOBase(self) -> local_io.BufferedIOBase:
        ...

    @property
    @abstractmethod
    def BufferedReader(self) -> local_io.BufferedReader:
        ...

    @property
    @abstractmethod
    def BufferedWriter(self) -> local_io.BufferedWriter:
        ...

    @property
    @abstractmethod
    def BufferedRWPair(self) -> local_io.BufferedRWPair:
        ...

    @property
    @abstractmethod
    def BufferedRandom(self) -> local_io.BufferedRandom:
        ...

    @property
    @abstractmethod
    def TextIOBase(self) -> local_io.TextIOBase:
        ...

    @property
    @abstractmethod
    def TextIOWrapper(self) -> local_io.TextIOWrapper:
        ...

    @property
    @abstractmethod
    def UnsupportedOperation(self) -> local_io.UnsupportedOperation:
        ...

    @property
    @abstractmethod
    def IncrementalNewlineDecoder(self) -> local_io.IncrementalNewlineDecoder:
        ...

    @property
    @abstractmethod
    def BlockingIOError(self) -> local_io.BlockingIOError:
        ...

    @property
    @abstractmethod
    def SEEK_SET(self):
        return self.local_io.SEEK_SET

    @property
    @abstractmethod
    def SEEK_CUR(self):
        return self.local_io.SEEK_CUR

    @property
    @abstractmethod
    def SEEK_END(self):
        return self.local_io.SEEK_END

    @property
    @abstractmethod
    def DEFAULT_BUFFER_SIZE(self):
        return self.local_io.DEFAULT_BUFFER_SIZE

    @abstractmethod
    def open(
        self,
        file,
        mode="r",
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None,
        closefd=True,
        opener=None,
    ) -> typing.Any:
        ...

    @abstractmethod
    def open_code(self, path: str):
        return self.local_io.open_code("")

    @abstractmethod
    def text_encoding(self, encoding, stacklevel=2) -> typing.Any | typing.Literal["utf-8", "locale"]:
        ...


class AbcIo(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mio"

    __all__ = _interface.local_io.__all__

    __doc__ = _interface.local_io.__doc__

    __delegated__ = {_interface: _interface.local_io}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_io, self, {"io": io, "msvcrt": msvcrt})

        _test_valid_size_type = list()
        BufferedReader = mod.BufferedReader

        class BufferedReaderFix(mod.BufferedReader):
            def read(self, size=None):
                _test_valid_size_type[size:]
                data = BufferedReader.read(self, size)
                return data

        mod.BufferedReader = BufferedReaderFix

        self.__delegate__(_interface, mod)



# -*- coding: utf-8 -*-
import unittest
import bmy

from myrrh.core.services.groups import MyrrhGroup


class TestBuild(unittest.TestCase):
    def setUp(self):
        bmy.init()

    def test_build1(self):
        e1 = bmy.new(path="**/local", eid="e1")
        self.assertEqual(bmy.current(e1), e1)
        bmy.build(eid=e1)
        self.assertTrue(bmy.isbuilt(eid=e1))
        e2 = bmy.new(path="**/local", eid="e2")
        bmy.build(eid=e2)
        self.assertTrue(bmy.isbuilt(eid=e2))

    def test_buildn(self):
        e1 = bmy.new(path="**/local", eid="e1")
        self.assertEqual(bmy.current(e1), e1)
        bmy.build(eid=(e1,))
        self.assertTrue(bmy.isbuilt(eid=e1))
        e2 = bmy.new(path="**/local", eid="e2")
        bmy.build(eid=(e2,))
        self.assertTrue(bmy.isbuilt(eid=e2))
        e3 = bmy.new(path="**/local", eid="e3")
        e4 = bmy.new(path="**/local", eid="e4")
        e34 = bmy.build(eid=(e3, e4))
        self.assertTrue(bmy.isbuilt(eid=e2))
        self.assertTrue(bmy.isbuilt(eid=e34))

    def test_build_eid(self):
        e1 = bmy.new(path="**/local", eid="e1")
        e2 = bmy.new(path="**/local", eid="e2")
        bmy.build(eid=e2)
        self.assertTrue(bmy.isbuilt(eid=e2))
        self.assertFalse(bmy.isbuilt(eid=e1))
        bmy.build(eid=e1)
        self.assertTrue(bmy.isbuilt(eid=e1))


class TestBmyEntity(unittest.TestCase):
    def test_basic_entity(self):
        eids = bmy.new(path="**/local", eid=("e1", "e2"))
        bmy.build(eid=eids)
        es = bmy.entity(eid=eids)

        self.assertIsInstance(es, MyrrhGroup)


class TestBmyNew(unittest.TestCase):
    def test_basic_new_setting(self):
        eid = bmy.new(path="**/local", cwd="doesnotexist", eid="e1")
        self.assertRaises(bmy.BmyMyrrhFailure, bmy.build, eid=eid)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from myrrh.framework.msh.madvsh import AbcAdvSh

_cmd_internal = (
    b"assoc",
    b"call",
    b"cd",
    b"cls",
    b"color",
    b"copy",
    b"date",
    b"del",
    b"dir",
    b"echo",
    b"endlocal",
    b"erase",
    b"exit",
    b"for",
    b"ftype",
    b"goto",
    b"if",
    b"md",
    b"mklink",
    b"move",
    b"path",
    b"pause",
    b"popd",
    b"prompt",
    b"pushd",
    b"rem",
    b"ren",
    b"rd",
    b"set",
    b"setlocal",
    b"shift",
    b"start",
    b"time",
    b"title",
    b"type",
    b"ver",
    b"verify",
    b"vol",
    b"::",
    b"(",
    b"@",
)
_cmd_operators = set((b"&&", b"|", b"||", b";", b";;", b"&", b">>", b"<<", b"<", b">", b">|", b">&", b"<&", b"<<-", b"<>", b"(", b")"))


__mlib__ = "AdvSh"


class AdvSh(AbcAdvSh):
    def initialize_stat(self):
        self.execute(b"%(ping)s -n 1 127.0.0.1" % self.myrrh_os.getbinb, count=1).values()

    def _calibration_execution(self, count):
        return self.execute(b"%(ping)s -n 2 127.0.0.1" % self.myrrh_os.getbinb, count=count)



from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcGzip"


class _interface(ABC):
    import gzip as local_gzip

    @property
    @abstractmethod
    def BadGzipFile(self) -> local_gzip.BadGzipFile:
        ...

    @property
    @abstractmethod
    def GzipFile(self) -> local_gzip.GzipFile:
        ...

    @property
    @abstractmethod
    def _COMPRESS_LEVEL_BEST(self):
        return self.local_gzip._COMPRESS_LEVEL_BEST

    @property
    @abstractmethod
    def _COMPRESS_LEVEL_FAST(self):
        return self.local_gzip._COMPRESS_LEVEL_FAST

    @property
    @abstractmethod
    def _COMPRESS_LEVEL_TRADEOFF(self):
        return self.local_gzip._COMPRESS_LEVEL_TRADEOFF

    @property
    @abstractmethod
    def READ(self):
        return self.local_gzip.READ

    @property
    @abstractmethod
    def WRITE(self):
        return self.local_gzip.WRITE

    @property
    @abstractmethod
    def FTEXT(self):
        return self.local_gzip.FTEXT

    @property
    @abstractmethod
    def FHCRC(self):
        return self.local_gzip.FHCRC

    @property
    @abstractmethod
    def FEXTRA(self):
        return self.local_gzip.FEXTRA

    @property
    @abstractmethod
    def FNAME(self):
        return self.local_gzip.FNAME

    @property
    @abstractmethod
    def FCOMMENT(self):
        return self.local_gzip.FCOMMENT

    @abstractmethod
    def open(
        self,
        filename,
        mode="rb",
        compresslevel=_COMPRESS_LEVEL_BEST,
        encoding=None,
        errors=None,
        newline=None,
    ) -> local_gzip.GzipFile:
        ...

    @abstractmethod
    def compress(self, data, compresslevel=_COMPRESS_LEVEL_BEST, *, mtime=None) -> bytes:
        ...

    @abstractmethod
    def decompress(self, data) -> bytes:
        ...


class AbcGzip(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mgzip"

    __all__ = ["BadGzipFile", "GzipFile", "open", "compress", "decompress"]

    __doc__ = _interface.local_gzip.__doc__

    __delegated__ = {_interface: _interface.local_gzip}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_gzip, self)

        self.__delegate__(_interface, mod)



# flake8: noqa: E722
import sys
import typing

from io import BytesIO

from ..managers import runtime_cached_property
from ..managers import RuntimeTaskManager, RuntimeObjectManager

from ...interfaces import IMyrrhOs, IRuntimeObject, IFileInStream, IFileOutStream
from ....provider import Wiring

from ..objects import (
    FileStream,
    FileInOutStream,
    FileInStreamAsync,
    FileOutStreamAsync,
    PipeInStreamAsync,
    PipeOutStreamAsync,
    FileInStream,
    FileOutStream,
    Process,
    InPipe,
    OutPipe,
    Buffer,
    Pipe,
    MHandle,
    InputIo,
)

from ...services.entity import Entity
from ...services import cfg_init
from ....provider import Protocol as _Protocol, Wiring as _Wiring


class RuntimeSyscall:
    MAX_FDS = cfg_init("max_fds", None, section="myrrh.core.services.system")
    POOL_SIZE = cfg_init("system_max_concurren_tasks", 10, section="myrrh.core.services.system")
    IDLE_TIMEOUT = cfg_init("runtime_task_idle_timeout", 0.01, section="myrrh.core.services.system")
    CHUNK_SIZE = cfg_init("runtime_stream_chunk_size", 65535, section="myrrh.core.services.system")

    Wiring = _Wiring
    Protocol = _Protocol

    def __init__(self, myrrh_os: IMyrrhOs):
        self.myrrh_os = myrrh_os

    @runtime_cached_property("fds", init_at_creation_time=True)
    def objects(self) -> RuntimeObjectManager:
        mngr = RuntimeObjectManager(self.MAX_FDS)

        self.hin = mngr.append(
            InputIo(),
            FileStream(sys.stdout.fileno(), b"", b"stdout", getattr(Entity, ".").system.stream),
            FileStream(sys.stderr.fileno(), b"", b"stderr", getattr(Entity, ".").system.stream),
        )
        self.hout = self.hin + 1
        self.herr = self.hin + 2

        return mngr

    @runtime_cached_property("tasks", init_at_creation_time=True)
    def tasks(self) -> RuntimeTaskManager:
        return RuntimeTaskManager(self.POOL_SIZE, self.myrrh_os.cfg.eid)

    def _a_(self, *obj):
        hint = self.objects.append(*obj)
        self.tasks.append(*obj)
        return hint

    def _make_obj_in(self, hint, filestream):
        handle = self.gethandle(self.dup(hint), False)
        if isinstance(handle, InPipe):
            task = PipeOutStreamAsync(filestream, handle)
        else:
            task = Pipe(handle, filestream)

        return task

    def _make_obj_out(self, hint, filestream):
        handle = self.gethandle(self.dup(hint), False)

        if isinstance(handle, OutPipe):
            task = PipeInStreamAsync(filestream, handle)
        else:
            task = Pipe(filestream, handle)

        return task

    def open_file(
        self,
        path: bytes,
        wiring: _Wiring,
        use_async=True,
        extras: dict | None = None,
        protocol: _Protocol | str | None = None,
    ) -> int:
        path = self.myrrh_os.p(path)
        stream = self.myrrh_os.Stream(protocol)

        path, handle = stream.open_file(path, wiring.value, extras=extras)

        filestream = None

        try:
            filestream = FileStream(
                handle,
                self.myrrh_os.dirname(path),
                self.myrrh_os.basename(path),
                self.myrrh_os.stream,
            )

            StreamCls = (FileInStream, FileOutStream) if use_async else (FileInStreamAsync, FileOutStreamAsync)

            match wiring & self.Wiring.INOUT:
                case self.Wiring.INOUT:
                    obj = FileInOutStream(filestream)
                case self.Wiring.IN:
                    obj = StreamCls[0](filestream)
                case self.Wiring.OUT:
                    obj = StreamCls[1](filestream)
                case _:
                    raise TypeError(f"open() invalid mode 0x{wiring.value:x}")

        except:
            try:
                if filestream:
                    filestream.close()
                else:
                    self.myrrh_os.stream.close(handle)
            except:
                pass
            raise

        try:
            return self._a_(obj)
        except:
            try:
                obj.close()
            except:
                pass

            raise

    def open_pipe(self) -> tuple[int, int]:
        buffer = Buffer()

        stream_rd = InPipe(buffer)
        stream_wr = OutPipe(buffer)

        hint = self._a_(stream_rd, stream_wr)
        return hint, hint + 1

    def open_process(
        self,
        path: bytes,
        args: list[bytes] = [],
        working_dir: bytes | None = None,
        env: dict | None = None,
        stdin: int | None = None,
        stdout: int | None = None,
        stderr: int | None = None,
        *,
        extras: dict | None = None,
        protocol: _Protocol | str | None = None,
    ) -> int:
        wiring = Wiring.OFF
        if stdin is not None:
            wiring |= Wiring.IN
        if stdout is not None:
            wiring |= Wiring.OUT
        if stderr is not None:
            wiring |= Wiring.ERR

        if path is None:
            raise ValueError("executable file path can not be None")

        stream = self.myrrh_os.Stream(protocol)
        path, hproc, hin, hout, herr = stream.open_process(path, wiring.value, args, working_dir, env, extras=extras)

        name = self.myrrh_os.basename(path)
        path = self.myrrh_os.dirname(path)

        if hin is not None:
            stdin = self._make_obj_in(stdin, FileStream(hin, path, name, stream))

        if hout is not None:
            stdout = self._make_obj_out(stdout, FileStream(hout, path, name, stream))

        if herr is not None:
            stderr = self._make_obj_out(stderr, FileStream(herr, path, name, stream))

        proc = Process(hproc, path, name, stream, hin, hout, herr)
        hproc = self._a_(proc)
        self.tasks.append(stdin, stdout, stderr)
        self.objects.register_proc(hproc)

        return hproc

    def stream_in(
        self,
        file_path: bytes,
        stream: BytesIO,
        *,
        extras: dict | None = None,
        protocol: _Protocol | str | None = None,
    ):
        fd = self.open_file(file_path, wiring=Wiring.IN, extras=extras, protocol=protocol)
        chunk_size = self.CHUNK_SIZE
        if extras:
            chunk_size = extras.get("chunk_size", chunk_size)

        with self.gethandle(fd) as handle:
            while buf := handle.read(chunk_size):
                sz = 0
                ln = len(buf)
                while sz < ln:
                    sz += stream.write(buf[sz:])

    def stream_out(
        self,
        file_path: bytes,
        stream: BytesIO,
        *,
        extras: dict | None = None,
        protocol: _Protocol | str | None = None,
    ):
        fd = self.open_file(
            file_path,
            wiring=Wiring.OUT | Wiring.CREATE | Wiring.RESET,
            extras=extras,
            protocol=protocol,
        )
        chunk_size = self.CHUNK_SIZE
        if extras:
            chunk_size = extras.get("chunk_size", chunk_size)

        with self.gethandle(fd) as handle:
            while buf := stream.read(chunk_size):
                sz = 0
                ln = len(buf)
                while sz < ln:
                    sz += handle.write(buf[sz:])

    def gethandle(self, hint: int, detach=False) -> MHandle:
        handle = self.objects.gethandle(hint)

        if detach:
            handle.detach()

        return handle

    def getallproc(self) -> tuple[Process]:
        return tuple(self.objects.procs.values())

    def getproc(self, pid: int) -> Process:
        try:
            return self.objects.getprocs(pid)
        except KeyError:
            pass

        raise ChildProcessError(f"unknown child process pid:{pid}")

    def waitobj(self, obj: IRuntimeObject, timeout: float | None = None) -> None:
        task = self.tasks.tget(obj.ehandle)
        if task:
            return task.join(timeout)

    def close(self, handle: MHandle | int) -> None:
        if isinstance(handle, MHandle):
            return handle.close()

        return self.objects.close(handle)

    def dup(self, handle: MHandle | int) -> int:
        return self.objects.dup(handle)

    def dup2(self, handle1: MHandle | int, handle2: MHandle | int) -> int:
        return self.objects.dup2(handle1, handle2)

    def wait(self, handle: int | MHandle, timeout: float | None = None) -> None:
        if not isinstance(handle, MHandle):
            handle = self.gethandle(handle, True)

        if handle.terminated() is None:
            return self.waitobj(handle, timeout)  # type: ignore[arg-type]

        return handle.terminated()

    def read(self, handle: int | MHandle, size, *, extras: dict | None = None):
        if not isinstance(handle, MHandle):
            handle = self.gethandle(handle, True)

        return handle.read(size, extras=extras)

    def write(self, handle: int, data: bytes, *, extras: dict | None = None):
        if not isinstance(handle, MHandle):
            handle = self.gethandle(handle, True)  # type: ignore[arg-type, assignment]

        return handle.write(data, extras=extras)  # type: ignore[attr-defined]



import typing

from . import mjson
from . import msecrets
from . import mcfg

cmds: list[typing.Callable] = []

cmds.extend((getattr(mjson, c) for c in mjson.__all__))
cmds.extend((getattr(msecrets, c) for c in msecrets.__all__))
cmds.extend((getattr(mcfg, c) for c in mcfg.__all__))



import abc
import threading
import typing

from ...warehouse.item import BaseItem
from ...provider import (
    IShellService,
    IFileSystemService,
    IStreamService,
    IStateService,
    ISnapService,
    IInstanceService,
    Protocol,
    IService,
)

__all__ = [
    "IServiceGroup",
    "IRegistry",
    "IServiceGroup",
    "ISystem",
    "IHost",
    "IVendor",
    "IEntity",
    "ICoreService",
    "ICoreStreamService",
    "ICoreShellService",
    "ICoreFileSystemService",
    "ICoreStateService",
    "ICoreSnapService",
    "ICoreInstanceService",
    "Protocol",
]


class ICoreService(IService):
    @classmethod
    @abc.abstractmethod
    def eref(cls) -> str:
        ...


class ICoreStreamService(IStreamService, ICoreService):
    __delegate_all__ = (IStreamService, ICoreService)


class ICoreFileSystemService(IFileSystemService, ICoreService):
    __delegate_all__ = (IFileSystemService, ICoreService)


class ICoreShellService(IShellService, ICoreService):
    __delegate_all__ = (IShellService, ICoreService)


class ICoreStateService(IStateService, ICoreService):
    __delegate_all__ = (IStateService, ICoreService)


class ICoreSnapService(ISnapService, ICoreService):
    __delegate_all__ = (ISnapService, ICoreService)


class ICoreInstanceService(IInstanceService, ICoreService):
    __delegate_all__ = (IInstanceService, ICoreService)


class IRegistry(dict, abc.ABC):
    @abc.abstractmethod
    def __getattr__(self, name) -> typing.Any:
        ...

    @abc.abstractmethod
    def predefined(self) -> list[typing.Any]:
        ...

    @abc.abstractmethod
    def defined_values(self) -> list[typing.Any]:
        ...

    @abc.abstractmethod
    def append(self, item: BaseItem, mode: typing.Literal["keep", "update", "replace"]):
        ...


class IServiceGroup(abc.ABC):
    @property
    @abc.abstractmethod
    def services(self) -> dict[str, typing.Type[ICoreService]]:
        ...

    @property
    @abc.abstractmethod
    def cfg(self) -> IRegistry:
        ...

    @property
    @abc.abstractmethod
    def lock(self) -> threading.RLock:
        ...


class ISystem(IServiceGroup):
    @property
    @abc.abstractmethod
    def shell(self) -> ICoreShellService:
        ...

    @property
    @abc.abstractmethod
    def fs(self) -> ICoreFileSystemService:
        ...

    @property
    @abc.abstractmethod
    def stream(self) -> ICoreStreamService:
        ...

    @abc.abstractmethod
    def Stream(self, protocol: str | Protocol | None = None) -> ICoreStreamService:
        ...

    @abc.abstractmethod
    def Fs(self, protocol: str | Protocol | None = None) -> ICoreFileSystemService:
        ...

    @abc.abstractmethod
    def Shell(self, protocol: str | Protocol | None = None) -> ICoreShellService:
        ...


class IHost(IServiceGroup):
    @property
    @abc.abstractmethod
    def state(self) -> ICoreStateService:
        ...

    @property
    @abc.abstractmethod
    def snap(self) -> ICoreSnapService:
        ...

    @property
    @abc.abstractmethod
    def inst(self) -> ICoreInstanceService:
        ...


class IVendor(IServiceGroup):
    ...


class IEntity:
    @property
    @abc.abstractmethod
    def system(self) -> ISystem:
        ...

    @property
    @abc.abstractmethod
    def host(self) -> IHost:
        ...

    @property
    @abc.abstractmethod
    def vendor(self) -> IVendor:
        ...

    @property
    @abc.abstractmethod
    def cfg(self) -> IRegistry:
        ...

    @property
    @abc.abstractmethod
    def eid(self) -> str:
        ...



copyright = """
Copyright (C) 2023 - pyanjel7
All Rights Reserved.
"""

license = (
    copyright
    + """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Also add information on how to contact you by electronic and paper mail.

  If the program does terminal interaction, make it output a short
notice like this when it starts in an interactive mode:

    <program>  Copyright (C) <year>  <name of author>
    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type `show c' for details.

The hypothetical commands `show w' and `show c' should show the appropriate
parts of the General Public License.  Of course, your program's commands
might be different; for a GUI interface, you would use an "about box".

  You should also get your employer (if you work as a programmer) or school,
if any, to sign a "copyright disclaimer" for the program, if necessary.
For more information on this, and how to apply and follow the GNU GPL, see
<https://www.gnu.org/licenses/>.

  The GNU General Public License does not permit incorporating your program
into proprietary programs.  If your program is a subroutine library, you
may consider it more useful to permit linking proprietary applications with
the library.  If this is what you want to do, use the GNU Lesser General
Public License instead of this License.  But first, please read
<https://www.gnu.org/licenses/why-not-lgpl.html>.

"""
)



# -*- coding: utf-8 -*-

"""
**Os environment module**
-------------------------
"""
import os
from myrrh.utils.mstring import typestr

from myrrh.core.interfaces import abstractmethod
from myrrh.core.services.system import AbcRuntime

__mlib__ = "AbcOsEnv"


class AbcOsEnv(AbcRuntime):
    __frameworkpath__ = "mpython._mosenv"

    __all__ = [
        "environ",
        "name",
        "linesep",
        "getenv",
        "putenv",
        "unsetenv",
        "gethome",
        "gettemp",
        "getlogin",
        "geteuid",
        "getegid",
        "getuid",
        "getgid",
        "getgroups",
        "device_encoding",
    ]

    uname_result = os.uname_result

    __all__.extend(["uname_result"])

    get_terminal_size = staticmethod(os.get_terminal_size)
    terminal_size = staticmethod(os.terminal_size)

    @property
    def linesep(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.linesepb)

    @property
    def environ(self):
        return self.myrrh_os.getenv()

    @environ.setter
    def environ(self, env):
        return self.myrrh_os.setenvb(env)

    @property
    def environb(self):
        return self.myrrh_os.getenvb()

    @environb.setter
    def environb(self, env):
        return self.myrrh_os.setenvb(env)

    @property
    @abstractmethod
    def name(self):
        ...

    def getenv(self, key, default=None):
        try:
            return self.environ[key]
        except KeyError:
            return self.myrrh_os.fsdecode(default)

    def getenvb(self, key, default=None):
        try:
            return self.environb[key]
        except KeyError:
            return self.myrrh_os.fsencode(default)

    def get_exec_path(self, env=None):
        if env and env.get(b"PATH") and env.get("PATH"):
            raise ValueError("env cannot contain 'PATH' and b'PATH' keys")

        if env is None:
            env = self.myrrh_os.getenvb()

        paths = env.get("PATH", env.get(b"PATH", self.myrrh_os.defpathb))

        return typestr(self.myrrh_os.fsencoding)([p for p in paths.split(self.myrrh_os.pathsepb)])

    @abstractmethod
    def putenv(self, key, value):
        ...

    @abstractmethod
    def unsetenv(self, key):
        ...

    @abstractmethod
    def gethome(self):
        ...

    @abstractmethod
    def gettemp(self):
        ...

    @abstractmethod
    def getlogin(self):
        ...

    @abstractmethod
    def geteuid(self):
        ...

    @abstractmethod
    def getegid(self):
        ...

    @abstractmethod
    def getuid(self):
        ...

    @abstractmethod
    def getgid(self):
        ...

    @abstractmethod
    def getgroups(self):
        ...

    def device_encoding(self, fd):
        # fd connected to a terminal unsupported => always None
        return None

    # Miscellaneous System Information

    @abstractmethod
    def cpu_count(self):
        ...



from myrrh.framework.mpython.mbuiltins import AbcBuiltins

__mlib__ = "Builtins"


class Builtins(AbcBuiltins):
    try:
        from builtins import WindowsError  # type: ignore[attr-defined]
    except ImportError:
        WindowsError = OSError



# -*- coding: utf-8 -*-

#
# Part of this code is inspirated from or is a direct copy of the python subprocess module
#
# The python subprocess module code is licenced under this terms:
#
# Copyright (c) 2003-2005 by Peter Astrand <astrand@lysator.liu.se>
#
# Licensed to PSF under a Contributor Agreement.
# See http://www.python.org/2.4/license for licensing details.

"""
**SubProcess management module**
-----------
"""
import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from myrrh.framework.mpython import mbuiltins

__mlib__ = "AbcSubProcess"


class _interface(ABC):
    import subprocess as local_subprocess

    @property
    @abstractmethod
    def _mswindows(self) -> bool:
        ...

    @property
    @abstractmethod
    def SubprocessError(self) -> local_subprocess.SubprocessError:
        ...

    @property
    @abstractmethod
    def CalledProcessError(self) -> local_subprocess.SubprocessError:
        ...

    @property
    @abstractmethod
    def TimeoutExpired(self) -> local_subprocess.SubprocessError:
        ...

    @property
    @abstractmethod
    def CompletedProcess(self) -> local_subprocess.CompletedProcess:
        ...

    @property
    @abstractmethod
    def Popen(self) -> local_subprocess.Popen:
        ...

    @property
    @abstractmethod
    def PIPE(self) -> int:
        ...

    @property
    @abstractmethod
    def DEVNULL(self) -> int:
        ...

    @property
    @abstractmethod
    def STDOUT(self) -> int:
        ...

    @property
    @abstractmethod
    def STARTF_USESHOWWINDOW(self) -> int:
        ...

    @property
    @abstractmethod
    def SW_HIDE(self) -> int:
        ...

    @property
    @abstractmethod
    def _USE_POSIX_SPAWN(self) -> bool:
        ...

    @property
    @abstractmethod
    def _active(self) -> list:
        ...

    @abstractmethod
    def _cleanup(self) -> None:
        ...

    @property
    @abstractmethod
    def STARTUPINFO(self) -> typing.Any:
        ...

    @abstractmethod
    def call(self, *popenargs, timeout=None, **kwargs) -> typing.Any:
        ...

    @abstractmethod
    def check_call(self, *popenargs, **kwargs) -> int:
        ...

    @abstractmethod
    def check_output(self, *popenargs, timeout=None, **kwargs) -> typing.Any:
        ...

    @abstractmethod
    def run(self, *popenargs, input=None, capture_output=False, timeout=None, check=False, **kwargs) -> local_subprocess.CompletedProcess:
        ...

    @abstractmethod
    def getoutput(self, cmd, *, encoding=None, errors=None) -> str:
        ...

    @abstractmethod
    def getstatusoutput(cmd, *, encoding=None, errors=None) -> tuple[int, str]:
        ...

    @abstractmethod
    def list2cmdline(self, seq: list) -> str:
        ...


class AbcSubProcess(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.msubprocess"

    __all__ = _interface.local_subprocess.__all__

    __delegated__ = {_interface: _interface.local_subprocess}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_subprocess, self)
        self.__delegate__(_interface, mod)

        mod._USE_POSIX_SPAWN = not mod._mswindows



import threading
import os
import stat

from ...services import cfg_init
from ...interfaces import IInOutStream, IInStream, IOutStream, ITask, ABCDelegation

from ._objects import RuntimeObject

PIPE_BUFFER_SIZE = cfg_init("pipe_buffer_size", 1024 * 1024, section="myrrh.core")
PIPE_CHUNK_SIZE = cfg_init("pipe_chunk_size", 2048, section="myrrh.core")

__all__ = ["Buffer", "Pipe"]


class _Lock:
    def __init__(self, lock: threading._RLock | None = None, name: str | None = None):
        self.name = name
        self.cond = threading.Condition(lock) if lock is not None else threading.Condition()

    def __enter__(self):
        self.acquire()

    def __exit__(self, *a, **kwa):
        self.release()

    def acquire(self, blocking=True, timeout=-1):
        return True if self.cond is None else self.cond.acquire(blocking, timeout)

    def release(self):
        if self.cond:
            self.cond.release()

    def wait(self, timeout=None):
        return True if self.cond is None else self.cond.wait(timeout)

    def wait_for(self, predicate, timeout=None):
        return predicate() if self.cond is None else self.cond.wait_for(predicate, timeout)

    def notify(self, n=1):
        if self.cond is not None:
            self.cond.notify(n)

    def notify_all(self):
        if self.cond is not None:
            self.cond.notify_all()


class Buffer(IInOutStream, RuntimeObject):
    def __init__(
        self,
        lock_wr=None,
        lock_rd=None,
        size=PIPE_BUFFER_SIZE,
        chunk_size=PIPE_CHUNK_SIZE,
    ):
        super().__init__(-1, b"", f"buffer-{id(self)}", None)

        self.wr_lock = _Lock(lock_wr, "write")
        self.rd_lock = _Lock(lock_rd, "read")

        self.max_buf_size = size
        self._chunk_size = chunk_size

        self.rd_size = 0
        self.wr_size = self.max_buf_size

        self.rd_pos = 0
        self.wr_pos = 0

        self.to_write_len = 0
        self.to_read_len = 0

        self._buffer = memoryview(bytearray(self.max_buf_size))
        self.closed = False

        self._eot = False
        self._writer_eot = {}

    def __str__(self):
        return f'Buffer(size={self.max_buf_size}){" -> closed" if self.closed else " -> eot" if self.eot else ""}'

    @property
    def eot(self):
        return self._eot

    def send_eot(self, eot, _id=None):
        with self.rd_lock, self.wr_lock:
            self._writer_eot[_id] = eot
            self._eot = all(self._writer_eot.values())

            self.rd_lock.notify_all()
            self.wr_lock.notify_all()

    def write(self, data, *, extras=None):
        timeout = extras.get("timeout") if extras else None

        if self.closed:
            raise BrokenPipeError(f"try to write on closed myrrh io buffer : {self.ehandle}")

        with self.wr_lock:
            while not self.wr_size and not self.closed and not self.eot:
                if not self.wr_lock.wait(timeout):
                    raise TimeoutError

            if self.closed or self.eot:
                raise BrokenPipeError

            self.to_write_len = min(len(data), self._chunk_size)
            self.to_write_len = min(self.to_write_len, self.wr_size)

            from_size_len = min(self.max_buf_size - self.wr_pos, self.to_write_len)

            self._buffer[self.wr_pos : from_size_len + self.wr_pos] = data[:+from_size_len]

            if self.to_write_len > from_size_len:
                self._buffer[0 : self.to_write_len - from_size_len] = data[from_size_len : self.to_write_len]

            self.wr_pos = (self.to_write_len + self.wr_pos) % self.max_buf_size
            self.wr_size -= self.to_write_len

        with self.rd_lock:
            if self.to_write_len:
                self.rd_size += self.to_write_len
                write_len = self.to_write_len
                self.to_write_len = 0
                self.rd_lock.notify_all()

        return write_len

    def flush(self, *, extras=None):
        with self.rd_lock, self.wr_lock:
            self.rd_size += self.to_write_len
            self.to_write_len = 0
            data = bytearray()
            to_read = self.rd_size
            while len(data) < to_read:
                data += self.read(self.rd_size)

        return data

    def sync(self, *, extras=None):
        return

    def read(self, nbytes=None, *, extras=None):
        timeout = extras.get("timeout") if extras else None

        data = memoryview(bytearray(self._chunk_size))

        with self.rd_lock:
            if nbytes == 0:
                return data.obj[:0]

            while not self.rd_size and not self.closed and not self.eot:
                if not self.rd_lock.wait(timeout=timeout):
                    raise TimeoutError

            if (self.closed or self.eot) and self.rd_size == 0:
                return data.obj[:0]

            self.to_read_len = min(self.rd_size, self._chunk_size)

            if nbytes is not None and not nbytes < 0:
                self.to_read_len = min(nbytes, self.to_read_len)

            from_pos_len = min(self.max_buf_size - self.rd_pos, self.to_read_len)
            data[:from_pos_len] = self._buffer[self.rd_pos : from_pos_len + self.rd_pos]

            if from_pos_len < self.to_read_len:
                data[from_pos_len : self.to_read_len] = self._buffer[0 : self.to_read_len - from_pos_len]

            self.rd_pos = (self.rd_pos + self.to_read_len) % self.max_buf_size
            self.rd_size -= self.to_read_len

        with self.wr_lock:
            if self.to_read_len:
                self.wr_size += self.to_read_len
                read_len = self.to_read_len
                self.to_read_len = 0
                self.wr_lock.notify_all()

        return data.obj[:read_len]

    def close(self, *, extras=None):
        with self.wr_lock:
            with self.rd_lock:
                if self.closed:
                    return

                self.closed = True

                self.wr_lock.notify_all()
                self.rd_lock.notify_all()

    def stat(self, *, extras=None):
        return os.stat_result((stat.S_IFIFO, 0, 0, 0, 0, 0, 0, 0, 0, 0))


class Pipe(ITask, ABCDelegation):
    def __init__(self, instream: IInStream, outstream: IOutStream):
        self.instream = instream
        self.outstream = outstream

        self.terminated_by_exception = False
        self.closed = False
        self.outstream.send_eot(False, id(self))

    def __str__(self):
        return f"pipe->{str(self.instream.eref)}, {str(self.outstream.eref)}"

    def task(self):
        try:
            data = self.instream.read(None)
            if data:
                self.outstream.write(data)
            else:
                self.outstream.send_eot(True, id(self))

        except Exception:
            self.terminated_by_exception = True
            self.outstream.send_eot(True, id(self))

            raise

    def terminated(self):
        return (self.terminated_by_exception and (self.instream.closed and self.outstream.closed)) or None



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcGlob"


class _interface(ABC):
    import glob as local_glob

    @abstractmethod
    def glob(self, pathname, *, root_dir=None, dir_fd=None, recursive=False, include_hidden=False) -> list[typing.AnyStr]:
        ...

    @abstractmethod
    def iglob(self, pathname, *, root_dir=None, dir_fd=None, recursive=False, include_hidden=False) -> typing.Iterator[typing.AnyStr]:
        ...

    @abstractmethod
    def has_magic(self, s) -> bool:
        ...

    @abstractmethod
    def escape(self, pathname) -> str:
        ...

    @abstractmethod
    def glob0(dirname, pattern) -> list:
        ...

    @abstractmethod
    def glob1(dirname, pattern) -> list:
        ...


class AbcGlob(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mglob"

    __all__ = ["glob", "iglob", "escape"]

    __doc__ = _interface.local_glob.__doc__

    __delegated__ = {_interface: _interface.local_glob}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_glob, self)

        self.__delegate__(_interface, mod)



import os
import sys

argv = sys.argv
while True:
    try:
        exit_code = os.spawnl(os.P_WAIT, sys.executable, sys.executable, "-m", "myrrh.tools.myrrhc.main", *argv[1:])
    except KeyboardInterrupt:
        exit_code = 0

    if exit_code:
        if len(sys.argv) > 1:
            break

        if exit_code != 2:
            print("Myrrhc crashed with exit_code: %s" % exit_code)
            print("restart y/n?")
            c = ""
            while c not in ("y", "Y", "n", "N"):
                c = sys.stdin.read(1)

            if c in ("y", "Y"):
                continue

        break

    sys.exit(exit_code)



import os
import click
import typing
import logging
import asyncio
import itertools

from traceback import format_exc
from click.core import MultiCommand


import bmy

from .exceptions import Abort, Exit, Failure, Reboot
from . import completion

from myrrh.core.services import (
    cfg_get,
    cfg_init,
    __version__,
    __license__,
    __copyright__,
    PID,
)
from myrrh.core.services.logging import log
from myrrh.core.services.groups import myrrh_group_keys

from prompt_toolkit.shortcuts import print_container
from prompt_toolkit.formatted_text import to_formatted_text, ANSI
from prompt_toolkit.widgets import Label
from prompt_toolkit.layout import VSplit, HSplit, Dimension as D
from prompt_toolkit.application.current import get_app_session


class EidOuput:
    def __init__(self, eid: str):
        self.output = get_app_session().output

        self.texts: dict[str, str] = dict()

        eid = bmy.current(eid)  # type: ignore[attr-defined]

        if eid and bmy.isgroup(eid):  # type: ignore[attr-defined]
            for e in bmy.groupkeys(eid):  # type: ignore[attr-defined]
                self.texts[e] = ""
        else:
            self.texts[eid] = ""

    def __enter__(self):
        return self

    def __getitem__(self, eid):
        return Output(self, eid)

    def _print_container(self):
        frames = list()
        bodies = dict()
        pages = list()

        output = get_app_session().output
        rows = output.get_size().rows

        for e, m in self.texts.items():
            if not m:
                continue

            header = e and Label(to_formatted_text([("class:bold italic", e)])) or None
            lines = m.splitlines()
            max_line_len = D(preferred=max(*map(len, lines)))

            bodies[e] = list()
            bodies[e].append(max_line_len)
            while lines:
                body = Label(
                    to_formatted_text(ANSI("\n".join(lines[: rows - 1]))),
                    width=max_line_len,
                )

                lines = lines[rows - 1 :]

                bodies[e].append(header and [header, body] or [body])
                header = None

        pages = list(itertools.zip_longest(*bodies.values()))

        if not pages:
            return

        lens = pages[0]
        pages = pages[1:]
        for p in pages:
            p = [*(([Label("", width=lens[i])] if v is None else v) for i, v in enumerate(p))]

            hsplit = list(map(HSplit, p))
            frames.append(VSplit(hsplit))

        for f in frames:
            print_container(f)

    def write(self, message, eid=None):
        if not message:
            return

        if eid is None:
            for k, v in self.texts.items():
                self.texts[k] = v + message

        elif bmy.isgroup(eid):
            for eid in bmy.groupkeys(eid):
                self.texts[eid] = self.texts.get(eid, "") + message
        else:
            self.texts[eid] = self.texts.get(eid, "") + message

    def __exit__(self, *a, **kwa):
        self.close()

    def close(self):
        self._print_container()


class Output:
    def __init__(self, writer: EidOuput, eid):
        self.writer = writer
        self.eid = eid

    def fileno(self):
        return self.writer.output.fileno()

    def flush(self):
        self.writer.output.flush()

    def write(self, data):
        self.writer.write(data, self.eid)

    def isatty(self):
        try:
            return self.writer.output.stdout.isatty()
        except AttributeError:
            return False


def secho(message=None, eid=None, output=None, nl=True, *args, **kwargs):
    write_output = output

    if nl:
        message = message and (message + os.linesep) or os.linesep

    if not write_output and bmy.isgroup(message):
        write_output = EidOuput(bmy.groupkeys(message))

    try:
        if write_output:
            if bmy.isgroup(message):
                for e, m in message._d_.items():
                    click.secho(m, write_output[e], False, *args, **kwargs)
            else:
                click.secho(message, write_output[eid], False, *args, **kwargs)
        else:
            click.secho(message, None, False, *args, **kwargs)
    finally:
        if not output and write_output:
            write_output.close()


def error(message=None, level=None, *args, **kwargs):
    if not level or log.isEnabledFor(level):
        secho(message.capitalize(), *args, **kwargs, err=True, fg="red")


def success(message=None, *args, **kwargs):
    secho(message.capitalize(), *args, **kwargs, fg="green")


def info(message=None, *args, **kwargs):
    secho(message, *args, **kwargs)


def print_table(message, cols, *args, **kwargs):
    line = ""

    count = 0
    for sz in cols:
        s = len(message) > count and message[count].replace("\n", "\\n").replace("\r", "\\r") or ""

        txt_sz = max(0, sz - 2) or sz
        expand_sz = sz + 4

        if len(s) >= txt_sz:
            s = s[0] + s[1 : txt_sz - 4] + "..." + s[-1]

        line = "".join([line, ("%s\t" % s).expandtabs(expand_sz)])

        count += 1

    info(line, *args, **kwargs)


def format_eid(eid=None, info=None):
    try:
        eid = bmy.current(eid)

        result = eid

        if bmy.isgroup(eid):
            val = myrrh_group_keys(eid)
            if len(val) == 1:
                result = val[0]
            else:
                result = ",".join(("[%s]" % e) if bmy.isbuilt(eid=e) else ("(%s)" % e) for e in eid)
                info = None

        if info.__class__.__name__ == "_bmy_async_result":
            info = info._d_(eid)

        bro, brc = ("[", "]") if bmy.isbuilt(eid=eid) else ("(", ")")
        return "".join((bro, ":".join(filter(None, (result, info))), brc))

    except bmy.BmyInvalidEid:
        return "(%s)" % (info or "myrrhc")


progressbar = click.progressbar
style = click.style


class Path(click.ParamType):
    pass


class Argument(click.Argument):
    pass


class Option(click.Option):
    pass


class Context(click.Context):
    def __init__(
        self,
        command: "Command",
        parent: typing.Self | None = None,
        info_name: str | None = None,
        obj: typing.Any = None,
        auto_envvar_prefix: str | None = None,
        default_map: dict[str, typing.Any] | None = None,
        terminal_width: int | None = None,
        max_content_width: int | None = None,
        resilient_parsing: bool = False,
        allow_extra_args: bool | None = None,
        allow_interspersed_args: bool | None = None,
        ignore_unknown_options: bool | None = None,
        help_option_names: list[str] | None = None,
        token_normalize_func: typing.Callable[[str], str] | None = None,
        color: bool | None = None,
        show_default: bool | None = None,
        *,
        cmdline: str = "",
        ident: int | None = None,
    ) -> None:
        super().__init__(
            command,
            parent=parent,
            info_name=info_name,
            obj=obj,
            auto_envvar_prefix=auto_envvar_prefix,
            default_map=default_map,
            terminal_width=terminal_width,
            max_content_width=max_content_width,
            resilient_parsing=resilient_parsing,
            allow_extra_args=allow_extra_args,
            allow_interspersed_args=allow_interspersed_args,
            ignore_unknown_options=ignore_unknown_options,
            help_option_names=help_option_names,
            token_normalize_func=token_normalize_func,
            color=color,
            show_default=show_default,
        )
        self.cmdline = cmdline
        self.ident = ident or parent and parent.ident  # type: ignore[has-type]
        self.help_option_names = ["-h", "--help"]

    def exit(self, code=0):
        raise Failure()


class Command(click.Command):
    context_class = Context

    @property
    def command_class(self):
        return Command

    def make_context(self, info_name: str | None, args: typing.List[str], parent: Context | None = None, **extra: typing.Any) -> Context:  # type: ignore[override]
        info_name = args[0] if args and not (isinstance(args, str) and args[0].startswith("-")) and (not parent or isinstance(parent.command, Group)) else None
        self.context_settings["cmdline"] = parent and parent.cmdline or " ".join(args)
        return super().make_context(info_name, args, parent=parent, **extra)  # type: ignore[return-value]


class Group(click.Group):
    context_class = Context
    command_class = Command
    ident: int | None = None

    @property
    def group_class(self):
        return Group

    def set_ident(self, ident):
        self.context_settings["ident"] = ident

    def invoke(self, ctx):
        try:
            return super().invoke(ctx)

        except (Abort, Exit, Reboot):
            raise

        except click.ClickException as e:
            error("[%s]: %s" % (ctx.command_path, e.format_message()))

        except bmy.BmyException as e:
            error(format_exc(), logging.DEBUG, eid=e.eid and str(e.eid))
            error("[%s]: %s" % (ctx.command_path, str(e)), eid=e.eid and str(e.eid))

        except Exception as e:
            error(format_exc(), logging.DEBUG)
            error(str(e))

        raise Failure

    def get_command(self, ctx, cmd_name):
        command = self.commands.get(cmd_name, self.commands.get("-default"))
        return command

    def make_context(self, info_name: str | None, args: typing.List[str], parent: Context | None = None, **extra: typing.Any) -> Context:  # type: ignore[override]
        info_name = args[0] if args and not args[0].startswith("-") and (not parent or isinstance(parent.command, Group)) else None
        self.context_settings["cmdline"] = parent and parent.cmdline or " ".join(args)
        return super().make_context(info_name, args, parent=parent, **extra)  # type: ignore[return-value]


class DefaultCommand(MultiCommand):
    context_class = Context
    command_class = Command
    ident: int | None = None

    def __init__(self, group, name="-default", *args, **kwargs):
        super().__init__(name, *args, no_args_is_help=False, invoke_without_command=True, **kwargs)
        self.group = group

    def get_command(self, ctx, cmd_name):
        return self.group.get_command(ctx, cmd_name)

    def list_commands(self, ctx: Context) -> typing.List[str]:  # type: ignore[override]
        return self.group.list_commands(ctx)

    def make_context(self, info_name: str | None, args: typing.List[str], parent: Context | None = None, **extra: typing.Any) -> Context:  # type: ignore[override]
        self.context_settings["cmdline"] = parent and parent.cmdline or " ".join(args)
        return super().make_context("", [], parent=parent, **extra)  # type: ignore[return-value]


def group(name=None, **attrs):
    return click.group(name=name, cls=Group, **attrs)


def command(name=None, **attrs):
    return click.command(name=name, cls=Command, **attrs)


def argument(*param_decls, **attrs):
    shell_complete = attrs.pop("shell_complete", None)

    if isinstance(shell_complete, (list, tuple)):
        return click.argument(
            *param_decls,
            shell_complete=lambda _ctx, _cli, _incomplet: shell_complete,
            **attrs,
            cls=Argument,
        )

    return click.argument(*param_decls, shell_complete=shell_complete, **attrs, cls=Argument)


def option(*param_decls, **attrs):
    return click.option(*param_decls, **attrs, cls=Option)


def pass_option_eid(func):
    return option(
        "--eid",
        multiple=True,
        default=None,
        type=click.STRING,
        shell_complete=completion.eids_completer,
        help="entity id",
    )(func)


def pass_option_wdir(func):
    return option(
        "-w",
        "--working_dir",
        default=None,
        type=click.STRING,
        shell_complete=completion.path_completer(),
        help="working directory",
    )(func)


def option_cfg_default(name, default_value, section="myrrh.tools.myrrhc"):
    cfg_init(name, default_value, section)

    def default():
        return cfg_get(name, default_value, section=section)

    return default


def option_cfg_get(name, default_value=None, section="myrrh.tools.myrrhc"):
    return cfg_get(name, default_value, section=section)


def option_cfg_init(name, default_value=None, section="myrrh.tools.myrrhc"):
    cfg_init(name, default_value, section=section)


def print_version(ctx, param, value):
    if not value:
        return
    info("%s" % __version__)
    if param:
        ctx.exit()


def verbose(ctx, param, value):
    if not value:
        return

    if value:
        level = [logging.CRITICAL, logging.ERROR, logging.INFO, logging.DEBUG][min(3, value - 1)]
        bmy.debug(level=level)


pass_context = click.pass_context
CommandCollection = click.CommandCollection


@group(invoke_without_command=True)
@option("-v", "--verbose", count=True, callback=verbose, expose_value=False, is_eager=True)
@option("--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True)
@click.help_option()
@pass_context
def myrrhc_cmds(ctx):
    pass


@myrrhc_cmds.command()
def license():
    """
    Display the full license text
    """
    info(__license__)


@myrrhc_cmds.command()
def copyright():
    """
    Display the full license text
    """
    info(__copyright__)


@myrrhc_cmds.command()
@pass_context
def interact(ctx):
    """
      Start an interactive shell. All subcommands are available in it.

    :param old_ctx: The current Click context.
    :param prompt_kwargs: Parameters passed to
        :py:func:`prompt_toolkit.shortcuts.prompt`.

      If stdin is not a TTY, no prompt will be printed, but only commands read
      from stdin.
    """
    from . import prompt

    loop = asyncio.get_event_loop()
    if not loop.is_running():
        loop.run_until_complete(prompt.PromptSession(ctx).run())


@myrrhc_cmds.command()
@pass_context
def version(ctx):
    "Quit the console"
    print_version(ctx, None, True)


@myrrhc_cmds.command()
@pass_context
@argument("commands", required=False, default=None, nargs=-1)
def help(ctx, commands):
    "Show help"
    cmd = myrrhc_cmds
    for command in commands:
        cmd = cmd.get_command(ctx, command)
    info(cmd.get_help(ctx))
    ctx.exit()


@myrrhc_cmds.command(hidden=True)
def cls():
    "Clear the console"
    click.clear()


@myrrhc_cmds.command()
def pid():
    "Return the pid of the current console"
    info(f"{PID}")



import threading
import unittest
import mlib
import time


class _MLibTest(unittest.TestCase):
    def setUp(self):
        import bmy

        e1 = bmy.new(path="**/local", eid="e1")
        e2 = bmy.new(path="**/local", eid="e2")
        e3 = bmy.new(path="**/local", eid="e3")
        e4 = bmy.new(path="**/local", eid="e4")
        bmy.build(eid=e1)
        bmy.build(eid=e2)
        bmy.build(eid=e3)
        bmy.build(eid=e4)
        self.e1 = bmy.entity(e1).system
        self.e2 = bmy.entity(e2).system
        self.e3 = bmy.entity(e3).system
        self.e4 = bmy.entity(e4).system


class BasicTest(_MLibTest):
    def test_import(self):
        import sys

        import mlib
        import mlib.py
        import mlib.fs

        self.assertIs(sys.modules["mlib"], mlib)
        self.assertIs(sys.modules["mlib.py"], mlib.py)
        self.assertIs(sys.modules["mlib.fs"], mlib.fs)

    def test_select1(self):
        with mlib.mlib_select(self.e1):
            from mlib.py import os as os1

        with mlib.mlib_select(self.e2):
            from mlib.py import os as os2

        self.assertIsNot(os1, os2)
        self.assertEqual(os1.myrrh_os.cfg.id, self.e1.cfg.id)
        self.assertEqual(os2.myrrh_os.cfg.id, self.e2.cfg.id)

        from mlib.py import os

        self.assertIsNot(os, os1)
        self.assertIsNot(os, os2)

        from myrrh.core.interfaces import ABCDelegationMeta

        self.assertIsInstance(os, ABCDelegationMeta)

    def test_select_in_select(self):
        with mlib.mlib_select(self.e1):
            from mlib.py import os as os1

            with mlib.mlib_select(self.e2):
                from mlib.py import os as os2

        self.assertIsNot(os1, os2)
        self.assertEqual(os1.myrrh_os.cfg.id, self.e1.cfg.id)
        self.assertEqual(os2.myrrh_os.cfg.id, self.e2.cfg.id)

        from mlib.py import os

        self.assertIsNot(os, os1)
        self.assertIsNot(os, os2)

        from myrrh.core.interfaces import ABCDelegationMeta

        self.assertIsInstance(os, ABCDelegationMeta)


class TestWrappedModule(_MLibTest):
    def validate(self, os1, os2, osh1, osh2):
        self.assertIsNot(os1, os2)
        self.assertEqual(os1.myrrh_os.cfg.id, self.e1.cfg.id)
        self.assertEqual(os2.myrrh_os.cfg.id, self.e2.cfg.id)

        self.assertIsNot(self.e1, self.e2)
        self.assertIsNot(osh1, osh2)

        self.assertNotEqual(self.e1.cfg.id, self.e2.cfg.id)
        self.assertEqual(osh1.myrrh_os.cfg.id, self.e1.cfg.id)
        self.assertEqual(osh2.myrrh_os.cfg.id, self.e2.cfg.id)

        self.assertIsNot(osh1._mod, osh2._mod)
        self.assertEqual(osh1._mod.os.myrrh_os.cfg.id, self.e1.cfg.id)
        self.assertEqual(osh2._mod.os.myrrh_os.cfg.id, self.e2.cfg.id)

        cwd1 = os1.getcwd()
        cwd2 = os2.getcwd()

        with osh1.change_cwd(".."):
            self.assertNotEqual(cwd1, os1.getcwd())
            self.assertEqual(cwd2, os2.getcwd())

        self.assertEqual(cwd1, os1.getcwd())

    def test_select_with_os_helper(self):
        with mlib.mlib_select(self.e1):
            from mlib.py import os as os1
            from mlib.py.test.support import os_helper as osh1

        with mlib.mlib_select(self.e2):
            from mlib.py import os as os2
            from mlib.py.test.support import os_helper as osh2

        self.validate(os1, os2, osh1, osh2)

    def test_in_select_with_os_helper(self):
        with mlib.mlib_select(self.e1):
            from mlib.py.test.support import os_helper as osh1

            with mlib.mlib_select(self.e2):
                from mlib.py import os as os2
                from mlib.py.test.support import os_helper as osh2

            from mlib.py import os as os1

        self.validate(os1, os2, osh1, osh2)


class TestThreaded(_MLibTest):
    def validate(self, os1, osh1, e):
        self.assertEqual(os1.myrrh_os.cfg.id, e.cfg.id)

        self.assertEqual(osh1.myrrh_os.cfg.id, e.cfg.id)

        self.assertEqual(osh1._mod.os.myrrh_os.cfg.id, e.cfg.id)

        cwd1 = os1.getcwd()

        with osh1.change_cwd(".."):
            self.assertNotEqual(cwd1, os1.getcwd())
        self.assertEqual(cwd1, os1.getcwd())

    def _th1(self):
        for i in range(1, 100):
            with mlib.mlib_select(self.e1):
                from mlib.py import os

                time.sleep(0.1)
                from mlib.py.test.support import os_helper

                with mlib.mlib_select(self.e3):
                    from mlib.py import os as os3

                    time.sleep(0.1)
                    from mlib.py.test.support import os_helper as os_helper3

            self.validate(os, os_helper, self.e1)
            self.validate(os3, os_helper3, self.e3)

    def _th2(self):
        for i in range(1, 500):
            with mlib.mlib_select(self.e2):
                from mlib.py import os
                from mlib.py.test.support import os_helper

            with mlib.mlib_select(self.e4):
                from mlib.py import os as os4
                from mlib.py.test.support import os_helper as os_helper4

            self.validate(os, os_helper, self.e2)
            self.validate(os4, os_helper4, self.e4)

    def test_threading(self):
        th1 = threading.Thread(target=self._th1)
        th2 = threading.Thread(target=self._th2)

        th1.start()
        th2.start()

        th1.join()
        th2.join()


if __name__ == "__main__":
    unittest.main()



import collections
import weakref
import threading

from functools import wraps

from myrrh.core.services import groups, cfg_init
from myrrh.core.services.logging import log

from ._bmy_exceptions import BmyInvalidEid, BmyEidInUsed
from ._bmy_entity import BmyEntity

__all__ = ["init"]


def _dispatch_on_eid_type(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        eid = kwargs.get("eid")

        if entities.isgroup(eid):
            kwargs["eid"] = groups.MyrrhGroup(keys=eid)
            return groups.myrrh_group(func)(*args, **kwargs)

        return func(*args, **kwargs)

    return wrapper


_histories: dict[int, object] = weakref.WeakValueDictionary()  # type: ignore[assignment]


def _get_thread_history(ident=None):
    ident = ident or threading.get_ident()
    return _histories.get(ident) or _BmyEntities._thread_locals.history


class _BmyHistoryCollection(collections.deque):
    def __init__(self, *a, **kwa):
        super().__init__(*a, **kwa)
        self.pos = 0
        self.n = 0
        _histories[threading.get_ident()] = self

    def inc(self):
        return min(self.pos + 1, self.n)

    def dec(self):
        return max(self.pos - 1, 0)

    def get_current(self):
        return self[self.pos]

    def get_next(self):
        return self[self.inc()]

    def get_previous(self):
        return self[self.dec()]

    def goto_next(self):
        self.pos = self.inc()
        return self.get_current()

    def goto_previous(self):
        self.pos = self.dec()
        return self.get_current()

    def push(self, value):
        self.n = self.pos + 1
        self.n = self.n % self.maxlen
        if not self[self.pos] is None:
            self.pos = self.inc()
        self[self.pos] = value


class _BmyEntitiesLocal(threading.local):
    _HISTORY_SZ = cfg_init("history_size", 20, section="myrrh.framework.bmy")

    def __init__(self):
        self.history = _BmyHistoryCollection((None,) * (self._HISTORY_SZ), maxlen=self._HISTORY_SZ)


class _BmyEntities:
    _entities: dict[str, BmyEntity] = {}
    _lock_entities = threading.RLock()
    _HISTORY_SZ = cfg_init("history_size", 20, section="myrrh.framework.bmy")

    _thread_locals = _BmyEntitiesLocal()

    def __init__(self, entities=[]):
        with self._lock_entities:
            for e in entities:
                self.append(e)

    @property
    def history(self):
        return self._thread_locals.history

    @property
    def eids(self):
        """
         Copy of current entity id list

        :type: list(str)
        """
        with self._lock_entities:
            return list(self._entities.keys())

    @property
    def entities(self):
        """
         Copy of current entity list

        :getter: return the current entity id list
        :type: list(:class:`myrrh.core.Entity`)
        """
        with self._lock_entities:
            return list(self._entities.values())

    @property
    def eid(self):
        """
         Default entity id

        :return: entity id
        :rtype: str
        """
        return self.history.get_current()

    @eid.setter
    def eid(self, eid):
        """
         Set the default entity eid

        :setter str eid: entity id

        :raises BmyInvalidEid: on unknown entity
        """
        if self.isgroup(eid):
            eid = groups.myrrh_group_keys(eid)

        with self._lock_entities:
            if self.eid:
                self.get(eid=self.eid).release()

            # history[i] = next eid

            self.history.push(str(eid) if isinstance(eid, BmyEntity) else eid)

            if self.eid:
                entities = self.get(eid=self.eid)
                entities.acquire()

        log.debug(f"bmy new entity selected {str(self.eid)}")

    @groups.myrrh_group_sync_member
    def eid_(self, ident=None):
        return _get_thread_history(ident).get_current()

    @groups.myrrh_group_sync_member
    def append(self, assembly, eid):
        bmy_entity = BmyEntity(assembly, eid)

        with self._lock_entities:
            self._entities[eid] = bmy_entity

            if len(self._entities) == 1 or self.eid == bmy_entity.eid:
                # select or reselect entity
                self.eid = bmy_entity.eid

        return bmy_entity.eid

    @_dispatch_on_eid_type
    @groups.myrrh_group_sync_member
    def remove(self, *, eid):
        with self._lock_entities:
            entity = self.get(eid)

            if entity.used:
                raise BmyEidInUsed(eid, msg='can not remove entity "%s", the entity is in used' % eid)

            self._entities.pop(eid, None)

    @_dispatch_on_eid_type
    @groups.myrrh_group_sync_member
    def get(self, *, eid) -> BmyEntity:
        with self._lock_entities:
            if eid is not None and eid not in self._entities:
                raise BmyInvalidEid(eid)

            if eid is None:
                eid = self.eid

            if eid is None:
                raise BmyInvalidEid(eid)

            return self._entities[eid]

    @groups.myrrh_group_sync_member
    def __getitem__(self, eid) -> BmyEntity:
        """
         x[eid] <=> x.__getitem__(eid)

        :param str eid: entity id
        :raises:class:`myrrh.framework.bmy.BmyInvalidEid`: on unknown entity
        """
        return self.get(eid=eid)

    @groups.myrrh_group_sync_member
    def current(self, eid=None, ident=None):
        if eid:
            if isinstance(eid, BmyEntity):
                return eid.eid

            try:
                self[eid]  # raises on invalid eid
            except BmyInvalidEid:
                return
            else:
                return eid

        if eid is None:
            eid = self.eid_(ident)
            if eid is not None and self.isgroup(eid) and len(eid) == 1:
                return groups.myrrh_group_keys(eid)[0]
            return eid

        if self.isgroup(eid):
            eid = self.eid_(ident)

            if not eid:
                return ()

            if not self.isgroup(eid):
                return groups.MyrrhGroup(eid, keys=(eid,))

            return eid

        return self.eid_(ident)

    @groups.myrrh_group_sync_member
    def isgroup(self, eid=None):
        if eid is None:
            eid = self.current(eid)  # avoid infinite recursion with current
        return groups.is_myrrh_group(eid) or isinstance(eid, (tuple, list))

    @groups.myrrh_group_sync_member
    def groupkeys(self, group):
        if self.isgroup(group):
            return groups.myrrh_group_keys(group)

        return (group,)

    @groups.myrrh_group_sync_member
    def groupvalues(self, group):
        if self.isgroup(group):
            return groups.myrrh_group_values(group)

        return (group,)


def init(_entities=[]):
    """
    Initialize/reset bmy entity manager

    Args:
        entities (dict): dictionary { id(str): entity(:class:`myrrh.core.entity.Entity`), ... } of already defined entities

    Note:
        this method is automatically called when the bmy module is loaded

    """
    global entities
    entities = _BmyEntities(_entities)


####

entities: _BmyEntities = None  # type: ignore[assignment]
init()



from ...interfaces import IProcess, Stat, ICoreStreamService
from ...services import cfg_init
from ....provider import StatField

from ._objects import RuntimeObject

POLL_STEP = cfg_init("process_poll_delay", 0.01, section="myrrh.core")

__all__ = ("Process",)


class Process(RuntimeObject, IProcess):
    pid = 0
    service: ICoreStreamService

    def __init__(
        self,
        handle: int,
        path: bytes,
        name: bytes,
        service: ICoreStreamService,
        hin: int | None,
        hout: int | None,
        herr: int | None,
    ):
        super().__init__(handle, path, name, service)

        if service.stat(handle, StatField.PID.value).get("st_pid") is None:
            raise RuntimeError(f'Unable to get pid of process: {path.decode(errors="ignore")}')

        self.pid = service.stat(handle, StatField.PID.value).get("st_pid")  # type: ignore[assignment]

        self._exit_status = None
        self.hs = (hin, hout, herr)

    def __str__(self):
        return f"{self.eref}->proc"

    def task(self):
        try:
            st = self.stat(StatField.STATUS)
            self._exit_status = st.st_status

            return self._exit_status

        except Exception:
            self._exit_status = -1
            raise

    @property
    def exit_status(self):
        return self._exit_status

    def terminate(self, *, extras=None):
        return self.service.terminate(self.ehandle, extras=extras)

    def stat(self, fields=StatField.ALL, *, extras=None) -> Stat:
        return Stat(**self.service.stat(self.ehandle, fields, extras=extras))

    def close(self):
        if not self.closed:
            self.closed = True
            self.service.close(self.ehandle)

    def terminated(self):
        return self._exit_status



import abc
import typing

from ._iservices import IService

__all__ = ["IProvider"]


class IProvider(abc.ABC):
    _name_: str  # reserved

    @abc.abstractmethod
    def services(self) -> tuple[typing.Type[IService]]:
        ...

    @abc.abstractmethod
    def catalog(self) -> tuple[str]:
        ...

    @abc.abstractmethod
    def deliver(self, name: str) -> dict | None:
        ...



import typing

import pydantic

from .item import BaseItem


class System(BaseItem[typing.Literal["system"]]):
    os: str = ""
    label: str = ""
    description: str = ""
    fsencoding: str = ""
    fsencodeerrors: str = ""
    encoding: str = ""
    encoding_errors: str = ""
    localecode: str = ""
    default_shell_encoding: str = ""

    max_size: int | None = None
    username: int | str | None = None
    uid: int | str | None = None
    gid: int | str | None = None
    groups: list[int] | None = pydantic.Field(default_factory=list)

    shell: str = pydantic.Field("")
    shellargs: tuple[str] = pydantic.Field(
        default_factory=list,
    )
    cwd: str = pydantic.Field(
        "",
    )
    tmpdir: str = pydantic.Field("")
    devnull: str = pydantic.Field("")
    defpath: str = pydantic.Field("")
    rdenv: list[str] = pydantic.Field(default_factory=list)
    env: dict[str, str] = pydantic.Field(default_factory=dict)
    bin: dict[str, str] = pydantic.Field(default_factory=dict)

    @pydantic.field_validator(
        "shell",
        "shellargs",
        "cwd",
        "tmpdir",
        "devnull",
        "defpath",
        "rdenv",
        "env",
        "bin",
        mode="before",
    )
    def decode_after(cls, value, info):
        encoding = info.data.get("encoding")
        encoding_errors = info.data.get("encoding_errors")
        return cls._decode(value, encoding, encoding_errors)

    @property
    def shellb(cls) -> str:
        return cls._encode(cls.shell, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def shellargsb(cls) -> list[str]:
        return cls._encode(cls.shellargs, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def cwdb(cls) -> bytes:
        return cls._encode(cls.cwd, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def tmpdirb(cls) -> bytes:
        return cls._encode(cls.tmpdir, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def devnullb(cls) -> bytes:
        return cls._encode(cls.devnull, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def defpathb(cls) -> bytes:
        return cls._encode(cls.defpath, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def rdenvb(cls) -> list[bytes]:
        return cls._encode(cls.rdenv, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def envb(cls) -> dict[bytes, bytes]:
        return cls._encode(cls.env, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @property
    def binb(cls) -> dict[bytes, bytes]:
        return cls._encode(cls.bin, cls.encoding, cls.encoding_errors)  # type: ignore[return-value]

    @classmethod
    def _encode(cls, value, encoding, encode_errors) -> bytes | list[bytes] | dict[bytes, bytes]:
        if isinstance(value, bytes):
            return value

        if isinstance(value, (list, tuple)):
            return [cls._encode(s, encoding, encode_errors) for s in value]  # type: ignore[misc]

        if isinstance(value, dict):
            return {cls._encode(k, encoding, encode_errors): cls._encode(v, encoding, encode_errors) for k, v in value.items()}  # type: ignore[misc]

        if isinstance(value, str):
            return value.encode(encoding or "utf8", errors=encode_errors)

        return value

    @classmethod
    def _decode(cls, value, encoding, encode_errors) -> str | list[str] | dict[str, str]:
        if isinstance(value, str):
            return value

        if isinstance(value, (list, tuple)):
            return [cls._decode(b, encoding, encode_errors) for b in value]  # type: ignore[misc]

        if isinstance(value, dict):
            return {cls._decode(k, encoding, encode_errors): cls._decode(v, encoding, encode_errors) for k, v in value.items()}  # type: ignore[misc]

        if isinstance(value, bytes):
            return value.decode(encoding or "utf8", errors=encode_errors)

        return value



import re
import os


from subprocess import list2cmdline as _list2cmdline

__all__ = ["split"]


# function copy from stackoverflow, Mar 9, 2016, author: kxr
# this function is distributed under the terms of CC BY-SA 3.0 licence.
# licence link : https://creativecommons.org/licenses/by-sa/3.0/
# updated to work with bytes argument
# original version : https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex/35900070#35900070
def _split(s, posix, RE_CMD_LEX, qs_replace, word_replace, empty_accu):
    args = []
    accu = None
    for qs, qss, esc, pipe, word, white, fail in re.findall(RE_CMD_LEX, s):
        if word:
            pass
        elif esc:
            word = esc[1]
        elif white or pipe:
            if accu is not None:
                args.append(accu)
            if pipe:
                args.append(pipe)
            accu = None
            continue
        elif fail:
            word = fail
        elif qs:
            word = qs_replace(qs)
            if posix:
                word = word_replace(word)
        else:
            word = qss  # may be even empty; must be last

        accu = (accu or empty_accu) + word

    if accu is not None:
        args.append(accu)

    return args


# function copy from stackoverflow, Mar 9, 2016, author: kxr
# this function is distributed under the terms of CC BY-SA 3.0 licence.
# licence link : https://creativecommons.org/licenses/by-sa/3.0/
# updated to work with bytes argument
# original version : https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex/35900070#35900070
def splitb(s, posix=(os.name == "posix")):
    """
    Multi-platform variant of shlex.split for command-line splitting
    For use with subprocess, for argv injection etc. Using fast REGEX
    """
    if posix:
        RE_CMD_LEX = rb""""((?:\\["\\]|[^"])*)"|'([^']*)'|(\\.)|(&&?|\|\|?|\d?>\>?|[<]<?)|([^\s'"\\&|<>]+)|(\s+)|(.)"""
    else:
        RE_CMD_LEX = rb""""((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?\>\>?|[<]<?)|([^\s"&|<>]+)|(\s+)|(.)"""

    def qs_replace(qs):
        return qs.replace(b'\\"', b'"').replace(b"\\\\", b"\\")

    def word_replace(word):
        return word.replace(b'""', b'"')

    return _split(s, posix, RE_CMD_LEX, qs_replace, word_replace, empty_accu=b"")


# function copy from stackoverflow, Mar 9, 2016, author: kxr
# this function is distributed under the terms of CC BY-SA 3.0 licence.
# licence link : https://creativecommons.org/licenses/by-sa/3.0/
# simplified version
# original version : https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex/35900070#35900070
def splits(s, posix=(os.name == "posix")):
    """
    Multi-platform variant of shlex.split for command-line splitting
    For use with subprocess, for argv injection etc. Using fast REGEX
    """
    if posix:
        RE_CMD_LEX = r""""((?:\\["\\]|[^"])*)"|'([^']*)'|(\\.)|(&&?|\|\|?|\d?>\>?|[<]<?)|([^\s'"\\&|<>]+)|(\s+)|(.)"""
    else:
        RE_CMD_LEX = r""""((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?\>\>?|[<]<?)|([^\s"&|<>]+)|(\s+)|(.)"""

    def qs_replace(qs):
        return qs.replace('\\"', '"').replace("\\\\", "\\")

    def word_replace(word):
        return word.replace('""', '"')

    return _split(s, posix, RE_CMD_LEX, qs_replace, word_replace, empty_accu="")


def split(s, posix=(os.name == "posix")):
    if isinstance(s, bytes):
        return splitb(s, posix)
    if isinstance(s, str):
        return splits(s, posix)

    raise TypeError('"s" must be of type str or bytes not "%s"' % type(s))


_meta_chars = '()%!^"<>&|'

_meta_re = re.compile("(" + "|".join(re.escape(char) for char in list(_meta_chars)) + ")")
_meta_map = {char: "^%s" % char for char in _meta_chars}

_meta_re_b = re.compile(b"(" + b"|".join(re.escape(char).encode() for char in list(_meta_chars)) + b")")
_meta_map_b = {char.encode(): b"^%s" % char.encode() for char in _meta_chars}
_meta_map_stringify_b = {char.encode(): b'"%s"' % char.encode() for char in _meta_chars}


# function copy from stackoverflow, Mar 23, 2015, author: Holder Just
# this function is distributed under the terms of CC BY-SA 3.0 licence.
# licence link : https://creativecommons.org/licenses/by-sa/3.0/
# updated to work with bytes argument
# original version : https://stackoverflow.com/questions/29213106/how-to-securely-escape-command-line-arguments-for-the-cmd-exe-shell-on-windows/29215357#29215357
def winshell_escape_for_cmd_exe_b(arg):
    # Escape an argument string to be suitable to be passed to cmd.exe on Windows
    #
    # This method takes an argument that is expected to already be properly
    # escaped for receiving program to be properly parsed. This argument
    # will be further escaped to pass the interpolation performed by cmd.exe
    # unchanged
    #
    # Any meta-characters will be escaped, removing the ability to e.g. use redirects or variables
    #
    # @program arg [String] a single command line argument to escape for cmd.exe
    # @return [String] an escaped string suitable to be passed as a program argument to cmd.exe

    return _meta_re_b.sub(lambda m: _meta_map_b[m.group(1)], arg)


# function copy from stackoverflow, Mar 23, 2015, author: Holder Just
# this function is distributed under the terms of CC BY-SA 3.0 licence.
# licence link : https://creativecommons.org/licenses/by-sa/3.0/
# updated to work with bytes argument
# original version : https://stackoverflow.com/questions/29213106/how-to-securely-escape-command-line-arguments-for-the-cmd-exe-shell-on-windows/29215357#29215357
def winshell_escape_for_stringyfication_in_cmdb(arg):
    return _meta_re_b.sub(lambda m: _meta_map_stringify_b[m.group(1)], arg)


def list2cmdlineb(seq, encode=os.fsencode, decode=os.fsdecode):
    if not len(seq):
        return b""

    if isinstance(seq[0], bytes):
        return encode(_list2cmdline(map(decode, seq)))

    return _list2cmdline(seq)


_find_unsafe = re.compile(rb"[^\w@%\+=:,\./\-<>]", re.ASCII).search


def quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return b"''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return b"'" + s.replace(b"'", b"'\"'\"'") + b"'"


def dquote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return b'""'
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return b'"' + s + b'"'



from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.framework.mpython.msys import AbcSys

__mlib__ = "Sys"


class Sys(AbcSys):
    __executable: bytes | None = None
    _abiflags = ""

    @property
    def _executable(self):
        rval = 1
        err = "python unavailable on android system"
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    @property
    def _platform(self):
        return "android"

    @property
    def abiflags(self):
        return self._get_sys_attr("abiflags")

    @property
    def _maxsize(self):
        return self.myrrh_os.fs.MAX_FILE_SIZE



import os
import textfsm

_dir = os.path.dirname(__file__)


def getparser(name):
    with open(os.path.join(_dir, name + ".textfsm")) as f:
        return Context(textfsm.TextFSM(f))


class Context:
    def __init__(self, textfsm):
        self.textfsm = textfsm

    def __enter__(self):
        return self.textfsm

    def __exit__(self, _t, _v, _tb):
        self.textfsm.Reset()



__all__ = [
    "BmyException",
    "BmyExecutionFailure",
    "BmyInvalidParameter",
    "BmyInvalidEid",
    "BmyMyrrhFailure",
    "BmyNotReady",
    "BmyTimeout",
]


class BmyException(Exception):
    def __init__(self, eid=None, func=None, msg="unexpected exception"):
        super().__init__(msg)
        self.eid = eid
        self.func = func


class BmyMyrrhFailure(BmyException):
    pass


class BmyExecutionFailure(BmyException):
    pass


class BmyInvalidParameter(BmyException):
    pass


class BmyInvalidEid(BmyInvalidParameter):
    def __init__(self, eid, func=None, msg=None):
        msg = msg or (eid and "invalid entity id: %s" % eid or "no entity id selected")
        super().__init__(eid, func, msg=msg)


class BmyEidInUsed(BmyInvalidParameter):
    def __init__(self, eid, func=None, msg=None):
        msg = msg or 'entity "%s" is in used' % eid
        super().__init__(eid, func, msg=msg)


class BmyNotReady(BmyException):
    def __init__(self, eid):
        msg = 'entity "%s" not ready, built is required' % eid
        super().__init__(msg=msg)
        self.eid = eid


class BmyTimeout(BmyException):
    def __init__(
        self,
        subject=None,
        reason=None,
        period=None,
        max_period=None,
        entity=None,
        msg=None,
    ):
        msg = msg or 'period expired for "%s", %s:%.3f>%.3f' % (
            subject,
            reason,
            period,
            max_period,
        )
        super().__init__(entity=entity, msg=msg)
        self.subject = subject
        self.reason = reason
        self.period = period
        self.max_period = max_period


class MyrrhCalibrationFailure(BmyException):
    pass



import pydantic
import uuid as _uuid
from myrrh.core.services import __version__
import typing

from .item import BaseItem

_entity = 0


def id_generator():
    global _entity
    _entity += 1
    return f"entity{_entity}"


class Id(BaseItem[typing.Literal["id"]]):
    id: str = pydantic.Field(default_factory=id_generator)
    uuid: _uuid.UUID = pydantic.Field(default=str(_uuid.UUID(int=0)))
    version: str = __version__

    def __str__(self):
        return str(self.id)



"""
**python msys system wrapper**

User should used this module (and avoid used of specific system functionality) to get cross-system portability
on their entity file operations.

To determinate the os depending implementation to use, this module checks the os field from the system item of
the entity

-----------
"""

from myrrh.core.services.system import AbcRuntime, _mlib_

from . import mimportlib

__mlib__ = "AbcSysConfig"


class AbcSysConfig(AbcRuntime):
    __frameworkpath__ = "mpython.msysconfig"

    __all__ = [
        "get_config_h_filename",
        "get_config_var",
        "get_config_vars",
        "get_makefile_filename",
        "get_path",
        "get_path_names",
        "get_paths",
        "get_platform",
        "get_python_version",
        "get_scheme_names",
        "parse_config_h",
    ]

    _INSTALL_SCHEMES = {
        "posix_prefix": {
            "stdlib": "{installed_base}/lib/python{py_version_short}",
            "platstdlib": "{platbase}/lib/python{py_version_short}",
            "purelib": "{base}/lib/python{py_version_short}/site-packages",
            "platlib": "{platbase}/lib/python{py_version_short}/site-packages",
            "include": "{installed_base}/include/python{py_version_short}{abiflags}",
            "platinclude": "{installed_platbase}/include/python{py_version_short}{abiflags}",
            "scripts": "{base}/bin",
            "data": "{base}",
        },
        "posix_home": {
            "stdlib": "{installed_base}/lib/python",
            "platstdlib": "{base}/lib/python",
            "purelib": "{base}/lib/python",
            "platlib": "{base}/lib/python",
            "include": "{installed_base}/include/python",
            "platinclude": "{installed_base}/include/python",
            "scripts": "{base}/bin",
            "data": "{base}",
        },
        "nt": {
            "stdlib": "{installed_base}/Lib",
            "platstdlib": "{base}/Lib",
            "purelib": "{base}/Lib/site-packages",
            "platlib": "{base}/Lib/site-packages",
            "include": "{installed_base}/Include",
            "platinclude": "{installed_base}/Include",
            "scripts": "{base}/Scripts",
            "data": "{base}",
        },
        "nt_user": {
            "stdlib": "{userbase}/Python{py_version_nodot}",
            "platstdlib": "{userbase}/Python{py_version_nodot}",
            "purelib": "{userbase}/Python{py_version_nodot}/site-packages",
            "platlib": "{userbase}/Python{py_version_nodot}/site-packages",
            "include": "{userbase}/Python{py_version_nodot}/Include",
            "scripts": "{userbase}/Python{py_version_nodot}/Scripts",
            "data": "{userbase}",
        },
        "posix_user": {
            "stdlib": "{userbase}/lib/python{py_version_short}",
            "platstdlib": "{userbase}/lib/python{py_version_short}",
            "purelib": "{userbase}/lib/python{py_version_short}/site-packages",
            "platlib": "{userbase}/lib/python{py_version_short}/site-packages",
            "include": "{userbase}/include/python{py_version_short}",
            "scripts": "{userbase}/bin",
            "data": "{userbase}",
        },
        "osx_framework_user": {
            "stdlib": "{userbase}/lib/python",
            "platstdlib": "{userbase}/lib/python",
            "purelib": "{userbase}/lib/python/site-packages",
            "platlib": "{userbase}/lib/python/site-packages",
            "include": "{userbase}/include",
            "scripts": "{userbase}/bin",
            "data": "{userbase}",
        },
    }

    _SCHEME_KEYS = (
        "stdlib",
        "platstdlib",
        "purelib",
        "platlib",
        "include",
        "scripts",
        "data",
    )

    os = mimportlib.module_property("os")
    io = mimportlib.module_property("io")
    sys = mimportlib.module_property("sys")

    def __init__(self, system=None):
        os = self.os
        sys = self.sys

        # FIXME don't rely on sys.version here, its format is an implementation detail
        # of CPython, use sys.version_info or sys.hexversion
        self._PY_VERSION = sys.version.split()[0]
        self._PY_VERSION_SHORT = "%d.%d" % sys.version_info[:2]
        self._PY_VERSION_SHORT_NO_DOT = "%d%d" % sys.version_info[:2]
        self._PREFIX = os.path.normpath(sys.prefix)
        self._BASE_PREFIX = os.path.normpath(sys.base_prefix)
        self._EXEC_PREFIX = os.path.normpath(sys.exec_prefix)
        self._BASE_EXEC_PREFIX = os.path.normpath(sys.base_exec_prefix)
        self._CONFIG_VARS = None
        self._USER_BASE = None
        self._PROJECT_BASE = self._safe_realpath(os.getcwd())
        if os.name == "nt" and self._PROJECT_BASE.lower().endswith(("\\pcbuild\\win32", "\\pcbuild\\amd64")):
            self._PROJECT_BASE = self._safe_realpath(os.path.join(self._PROJECT_BASE, os.path.pardir, os.path.pardir))

        # set for cross builds
        if "_PYTHON_PROJECT_BASE" in os.environ:
            self._PROJECT_BASE = self._safe_realpath(os.environ["_PYTHON_PROJECT_BASE"])

        self._sys_home = getattr(sys, "_home", None)
        if self._sys_home and os.name == "nt" and self._sys_home.lower().endswith(("\\pcbuild\\win32", "\\pcbuild\\amd64")):
            self._sys_home = os.path.dirname(os.path.dirname(self._sys_home))

        self._PYTHON_BUILD = self.is_python_build(True)

        if self._PYTHON_BUILD:
            for scheme in ("posix_prefix", "posix_home"):
                self._INSTALL_SCHEMES[scheme]["include"] = "{srcdir}/Include"
                self._INSTALL_SCHEMES[scheme]["platinclude"] = "{projectbase}/."

        self.__default_scheme = self._get_default_scheme()

    def _safe_realpath(self, path):
        realpath = self.os.path.realpath
        try:
            return realpath(path)
        except OSError:
            return path

    def _is_python_source_dir(self, d):
        os = self.os
        for fn in ("Setup.dist", "Setup.local"):
            if os.path.isfile(os.path.join(d, "Modules", fn)):
                return True
        return False

    def is_python_build(self, check_home=False):
        if check_home and self._sys_home:
            return self._is_python_source_dir(self._sys_home)
        return self._is_python_source_dir(self._PROJECT_BASE)

    def _subst_vars(self, s, local_vars):
        os = self.os
        try:
            return s.format(**local_vars)
        except KeyError:
            try:
                return s.format(**os.environ)
            except KeyError as var:
                raise AttributeError("{%s}" % var)

    def _extend_dict(self, target_dict, other_dict):
        target_keys = target_dict.keys()
        for key, value in other_dict.items():
            if key in target_keys:
                continue
            target_dict[key] = value

    def _expand_vars(self, scheme, vars):
        os = self.os
        res = {}
        if vars is None:
            vars = {}
        self._extend_dict(vars, self.get_config_vars())

        for key, value in self._INSTALL_SCHEMES[scheme].items():
            if os.name in ("posix", "nt"):
                value = os.path.expanduser(value)
            res[key] = os.path.normpath(self._subst_vars(value, vars))
        return res

    def _get_default_scheme(self):
        os = self.os
        if os.name == "posix":
            # the default scheme for posix is posix_prefix
            return "posix_prefix"
        return os.name

    def _getuserbase(self):
        os = self.os
        sys = self.sys
        env_base = os.environ.get("PYTHONUSERBASE", None)

        def joinuser(*args):
            return os.path.expanduser(os.path.join(*args))

        if os.name == "nt":
            base = os.environ.get("APPDATA") or "~"
            if env_base:
                return env_base
            else:
                return joinuser(base, "Python")

        if sys.platform == "darwin":
            framework = self.get_config_var("PYTHONFRAMEWORK")
            if framework:
                if env_base:
                    return env_base
                else:
                    return joinuser("~", "Library", framework, "%d.%d" % sys.version_info[:2])

        if env_base:
            return env_base
        else:
            return joinuser("~", ".local")

    def _parse_makefile(self, filename, vars=None):
        """Parse a Makefile-style file.

        A dictionary containing name/value pairs is returned.  If an
        optional dictionary is passed in as the second argument, it is
        used instead of a new dictionary.
        """
        # Regexes needed for parsing Makefile (and similar syntaxes,
        # like old-style Setup files).
        open = self.io.open
        os = self.os
        import re

        _variable_rx = re.compile(r"([a-zA-Z][a-zA-Z0-9_]+)\s*=\s*(.*)")
        _findvar1_rx = re.compile(r"\$\(([A-Za-z][A-Za-z0-9_]*)\)")
        _findvar2_rx = re.compile(r"\${([A-Za-z][A-Za-z0-9_]*)}")

        if vars is None:
            vars = {}
        done = {}
        notdone = {}

        with open(filename, errors="surrogateescape") as f:
            lines = f.readlines()

        for line in lines:
            if line.startswith("#") or line.strip() == "":
                continue
            m = _variable_rx.match(line)
            if m:
                n, v = m.group(1, 2)
                v = v.strip()
                # `$$' is a literal `$' in make
                tmpv = v.replace("$$", "")

                if "$" in tmpv:
                    notdone[n] = v
                else:
                    try:
                        v = int(v)
                    except ValueError:
                        # insert literal `$'
                        done[n] = v.replace("$$", "$")
                    else:
                        done[n] = v

        # do variable interpolation here
        variables = list(notdone.keys())

        # Variables with a 'PY_' prefix in the makefile. These need to
        # be made available without that prefix through sysconfig.
        # Special care is needed to ensure that variable expansion works, even
        # if the expansion uses the name without a prefix.
        renamed_variables = ("CFLAGS", "LDFLAGS", "CPPFLAGS")

        while len(variables) > 0:
            for name in tuple(variables):
                value = notdone[name]
                m1 = _findvar1_rx.search(value)
                m2 = _findvar2_rx.search(value)
                if m1 and m2:
                    m = m1 if m1.start() < m2.start() else m2
                else:
                    m = m1 if m1 else m2
                if m is not None:
                    n = m.group(1)
                    found = True
                    if n in done:
                        item = str(done[n])
                    elif n in notdone:
                        # get it on a subsequent round
                        found = False
                    elif n in os.environ:
                        # do it like make: fall back to environment
                        item = os.environ[n]

                    elif n in renamed_variables:
                        if name.startswith("PY_") and name[3:] in renamed_variables:
                            item = ""

                        elif "PY_" + n in notdone:
                            found = False

                        else:
                            item = str(done["PY_" + n])

                    else:
                        done[n] = item = ""

                    if found:
                        after = value[m.end() :]
                        value = value[: m.start()] + item + after
                        if "$" in after:
                            notdone[name] = value
                        else:
                            try:
                                value = int(value)
                            except ValueError:
                                done[name] = value.strip()
                            else:
                                done[name] = value
                            variables.remove(name)

                            if name.startswith("PY_") and name[3:] in renamed_variables:
                                name = name[3:]
                                if name not in done:
                                    done[name] = value

                else:
                    # bogus variable reference (e.g. "prefix=$/opt/python");
                    # just drop it since we can't deal
                    done[name] = value
                    variables.remove(name)

        # strip spurious spaces
        for k, v in done.items():
            if isinstance(v, str):
                done[k] = v.strip()

        # save the results in the global dictionary
        vars.update(done)
        return vars

    def get_makefile_filename(self):
        """Return the path of the Makefile."""
        sys = self.sys
        os = self.os
        if self._PYTHON_BUILD:
            return os.path.join(self._sys_home or self._PROJECT_BASE, "Makefile")
        if hasattr(sys, "abiflags"):
            config_dir_name = "config-%s%s" % (self._PY_VERSION_SHORT, sys.abiflags)
        else:
            config_dir_name = "config"
        if hasattr(sys.implementation, "_multiarch"):
            config_dir_name += "-%s" % sys.implementation._multiarch
        return os.path.join(self.get_path("stdlib"), config_dir_name, "Makefile")

    def _get_sysconfigdata_name(self):
        sys = self.sys
        os = self.os
        return os.environ.get(
            "_PYTHON_SYSCONFIGDATA_NAME",
            "_sysconfigdata_{abi}_{platform}_{multiarch}".format(
                abi=sys.abiflags,
                platform=sys.platform,
                multiarch=getattr(sys.implementation, "_multiarch", ""),
            ),
        )

    def _generate_posix_vars(self):
        """Generate the Python module containing build-time variables."""
        import pprint

        sys = self.sys
        os = self.os
        vars = {}
        # load the installed Makefile:
        makefile = self.get_makefile_filename()
        try:
            self._parse_makefile(makefile, vars)
        except OSError as e:
            msg = "invalid Python installation: unable to open %s" % makefile
            if hasattr(e, "strerror"):
                msg = msg + " (%s)" % e.strerror
            raise OSError(msg)
        # load the installed pyconfig.h:
        config_h = self.get_config_h_filename()
        try:
            with open(config_h) as f:
                self.parse_config_h(f, vars)
        except OSError as e:
            msg = "invalid Python installation: unable to open %s" % config_h
            if hasattr(e, "strerror"):
                msg = msg + " (%s)" % e.strerror
            raise OSError(msg)
        # On AIX, there are wrong paths to the linker scripts in the Makefile
        # -- these paths are relative to the Python source, but when installed
        # the scripts are in another directory.
        if self._PYTHON_BUILD:
            vars["BLDSHARED"] = vars["LDSHARED"]

        # There's a chicken-and-egg situation on OS X with regards to the
        # _sysconfigdata module after the changes introduced by #15298:
        # get_config_vars() is called by get_platform() as part of the
        # `make pybuilddir.txt` target -- which is a precursor to the
        # _sysconfigdata.py module being constructed.  Unfortunately,
        # get_config_vars() eventually calls _init_posix(), which attempts
        # to import _sysconfigdata, which we won't have built yet.  In order
        # for _init_posix() to work, if we're on Darwin, just mock up the
        # _sysconfigdata module manually and populate it with the build vars.
        # This is more than sufficient for ensuring the subsequent call to
        # get_platform() succeeds.
        name = self._get_sysconfigdata_name()
        if "darwin" in sys.platform:
            import types

            module = types.ModuleType(name)
            module.build_time_vars = vars
            sys.modules[name] = module

        pybuilddir = "build/lib.%s-%s" % (self.get_platform(), self._PY_VERSION_SHORT)
        if hasattr(sys, "gettotalrefcount"):
            pybuilddir += "-pydebug"
        os.makedirs(pybuilddir, exist_ok=True)
        destfile = os.path.join(pybuilddir, name + ".py")

        with open(destfile, "w", encoding="utf8") as f:
            f.write("# system configuration generated and used by" " the sysconfig module\n")
            f.write("build_time_vars = ")
            pprint.pprint(vars, stream=f)

        # Create file used for sys.path fixup -- see Modules/getpathb.c
        with open("pybuilddir.txt", "w", encoding="ascii") as f:
            f.write(pybuilddir)

    def _init_posix(self, vars):
        """Initialize the module as appropriate for POSIX systems."""
        # _sysconfigdata is generated at build time, see _generate_posix_vars()
        # name = self._get_sysconfigdata_name()
        # _temp = __import__(name, globals(), locals(), ['build_time_vars'], 0)
        # build_time_vars = _temp.build_time_vars
        # vars.update(build_time_vars)

    def _init_non_posix(self, vars):
        """Initialize the module as appropriate for NT"""
        os = self.os
        sys = self.sys
        get_path = self.get_path
        _PY_VERSION_SHORT_NO_DOT = self._PY_VERSION_SHORT_NO_DOT
        _safe_realpath = self._safe_realpath
        # set basic install directories
        vars["LIBDEST"] = get_path("stdlib")
        vars["BINLIBDEST"] = get_path("platstdlib")
        vars["INCLUDEPY"] = get_path("include")
        vars["EXT_SUFFIX"] = ".pyd"
        vars["EXE"] = ".exe"
        vars["VERSION"] = _PY_VERSION_SHORT_NO_DOT
        try:
            executable = sys._executable
        except AttributeError:
            executable = "no_python_on target"
        vars["BINDIR"] = os.path.dirname(_safe_realpath(executable))

    #
    # public APIs
    #

    def parse_config_h(self, fp, vars=None):
        """Parse a config.h-style file.

        A dictionary containing name/value pairs is returned.  If an
        optional dictionary is passed in as the second argument, it is
        used instead of a new dictionary.
        """
        if vars is None:
            vars = {}
        import re

        define_rx = re.compile("#define ([A-Z][A-Za-z0-9_]+) (.*)\n")
        undef_rx = re.compile("/[*] #undef ([A-Z][A-Za-z0-9_]+) [*]/\n")

        while True:
            line = fp.readline()
            if not line:
                break
            m = define_rx.match(line)
            if m:
                n, v = m.group(1, 2)
                try:
                    v = int(v)
                except ValueError:
                    pass
                vars[n] = v
            else:
                m = undef_rx.match(line)
                if m:
                    vars[m.group(1)] = 0
        return vars

    def get_config_h_filename(self):
        """Return the path of pyconfig.h."""
        os = self.os
        _PYTHON_BUILD = self._PYTHON_BUILD
        _sys_home = self._sys_home
        _PROJECT_BASE = self._PROJECT_BASE
        get_path = self.get_path
        if _PYTHON_BUILD:
            if os.name == "nt":
                inc_dir = os.path.join(_sys_home or _PROJECT_BASE, "PC")
            else:
                inc_dir = _sys_home or _PROJECT_BASE
        else:
            inc_dir = get_path("platinclude")
        return os.path.join(inc_dir, "pyconfig.h")

    def get_scheme_names(self):
        _INSTALL_SCHEMES = self._INSTALL_SCHEMES
        """Return a tuple containing the schemes names."""
        return tuple(sorted(_INSTALL_SCHEMES))

    def get_path_names(self):
        """Return a tuple containing the paths names."""
        return self._SCHEME_KEYS

    def get_paths(self, scheme=None, vars=None, expand=True):
        """Return a mapping containing an install scheme.

        ``scheme`` is the install scheme name. If not provided, it will
        return the default scheme for the current platform.
        """
        if scheme is None:
            scheme = self.__default_scheme

        if expand:
            return self._expand_vars(scheme, vars)
        else:
            return self._INSTALL_SCHEMES[scheme]

    def get_path(self, name, scheme=None, vars=None, expand=True):
        """Return a path corresponding to the scheme.

        ``scheme`` is the install scheme name.
        """
        if scheme is None:
            scheme = self.__default_scheme
        return self.get_paths(scheme, vars, expand)[name]

    def get_config_vars(self, *args):
        """With no arguments, return a dictionary of all configuration
        variables relevant for the current platform.

        On Unix, this means every variable defined in Python's installed Makefile;
        On Windows it's a much smaller set.

        With arguments, return a list of values that result from looking up
        each argument in the configuration variable dictionary.
        """
        _CONFIG_VARS = self._CONFIG_VARS
        sys = self.sys
        os = self.os
        if _CONFIG_VARS is None:
            self._CONFIG_VARS = {}
            _CONFIG_VARS = self._CONFIG_VARS
            # Normalized versions of prefix and exec_prefix are handy to have;
            # in fact, these are the standard versions used most places in the
            # Distutils.
            _CONFIG_VARS["prefix"] = self._PREFIX
            _CONFIG_VARS["exec_prefix"] = self._EXEC_PREFIX
            _CONFIG_VARS["py_version"] = self._PY_VERSION
            _CONFIG_VARS["py_version_short"] = self._PY_VERSION_SHORT
            _CONFIG_VARS["py_version_nodot"] = self._PY_VERSION_SHORT_NO_DOT
            _CONFIG_VARS["installed_base"] = self._BASE_PREFIX
            _CONFIG_VARS["base"] = self._PREFIX
            _CONFIG_VARS["installed_platbase"] = self._BASE_EXEC_PREFIX
            _CONFIG_VARS["platbase"] = self._EXEC_PREFIX
            _CONFIG_VARS["projectbase"] = self._PROJECT_BASE
            try:
                _CONFIG_VARS["abiflags"] = sys.abiflags
            except AttributeError:
                # sys.abiflags may not be defined on all platforms.
                _CONFIG_VARS["abiflags"] = ""

            if os.name == "nt":
                self._init_non_posix(_CONFIG_VARS)
            if os.name == "posix":
                self._init_posix(_CONFIG_VARS)
            # For backward compatibility, see issue19555
            SO = _CONFIG_VARS.get("EXT_SUFFIX")
            if SO is not None:
                _CONFIG_VARS["SO"] = SO
            # Setting 'userbase' is done below the call to the
            # init function to enable using 'get_config_var' in
            # the init-function.
            _CONFIG_VARS["userbase"] = self._getuserbase()

            # Always convert srcdir to an absolute path
            srcdir = _CONFIG_VARS.get("srcdir", self._PROJECT_BASE)
            if os.name == "posix":
                if self._PYTHON_BUILD:
                    # If srcdir is a relative path (typically '.' or '..')
                    # then it should be interpreted relative to the directory
                    # containing Makefile.
                    base = os.path.dirname(self.get_makefile_filename())
                    srcdir = os.path.join(base, srcdir)
                else:
                    # srcdir is not meaningful since the installation is
                    # spread about the filesystem.  We choose the
                    # directory containing the Makefile since we know it
                    # exists.
                    srcdir = os.path.dirname(self.get_makefile_filename())
            _CONFIG_VARS["srcdir"] = self._safe_realpath(srcdir)

            # OS X platforms require special customization to handle
            # multi-architecture, multi-os-version installers
            if sys.platform == "darwin":
                from . import _mosx_support

                _osx_support = _mlib_(_mosx_support)(self)
                _osx_support.customize_config_vars(_CONFIG_VARS)

        if args:
            vals = []
            for name in args:
                vals.append(_CONFIG_VARS.get(name))
            return vals
        else:
            return _CONFIG_VARS

    def get_config_var(self, name):
        """Return the value of a single variable using the dictionary returned by
        'get_config_vars()'.

        Equivalent to get_config_vars().get(name)
        """
        if name == "SO":
            import warnings

            warnings.warn("SO is deprecated, use EXT_SUFFIX", DeprecationWarning, 2)
        return self.get_config_vars().get(name)

    def get_platform(self):
        """Return a string that identifies the current platform.

        This is used mainly to distinguish platform-specific build directories and
        platform-specific built distributions.  Typically includes the OS name
        and version and the architecture (as supplied by 'os.uname()'),
        although the exact information included depends on the OS; eg. for IRIX
        the architecture isn't particularly important (IRIX only runs on SGI
        hardware), but for Linux the kernel version isn't particularly
        important.

        Examples of returned values:
           linux-i586
           linux-alpha (?)
           solaris-2.6-sun4u
           irix-5.3
           irix64-6.2

        Windows will return one of:
           win-amd64 (64bit Windows on AMD64 (aka x86_64, Intel64, EM64T, etc)
           win-ia64 (64bit Windows on Itanium)
           win32 (all others - specifically, sys.platform is returned)

        For other non-POSIX platforms, currently just returns 'sys.platform'.
        """
        sys = self.sys
        os = self.os
        if os.name == "nt":
            # sniff sys.version for architecture.
            prefix = " bit ("
            i = sys.version.find(prefix)
            if i == -1:
                return sys.platform
            j = sys.version.find(")", i)
            look = sys.version[i + len(prefix) : j].lower()
            if look == "amd64":
                return "win-amd64"
            if look == "itanium":
                return "win-ia64"
            return sys.platform

        if os.name != "posix" or not hasattr(os, "uname"):
            # XXX what about the architecture? NT is Intel or Alpha
            return sys.platform

        # Set for cross builds explicitly
        if "_PYTHON_HOST_PLATFORM" in os.environ:
            return os.environ["_PYTHON_HOST_PLATFORM"]

        # Try to distinguish various flavours of Unix
        osname, host, release, version, machine = os.uname()

        # Convert the OS name to lowercase, remove '/' characters
        # (to accommodate BSD/OS), and translate spaces (for "Power Macintosh")
        osname = osname.lower().replace("/", "")
        machine = machine.replace(" ", "_")
        machine = machine.replace("/", "-")

        if osname[:5] == "linux":
            # At least on Linux/Intel, 'machine' is the processor --
            # i386, etc.
            # XXX what about Alpha, SPARC, etc?
            return "%s-%s" % (osname, machine)
        elif osname[:5] == "sunos":
            if release[0] >= "5":  # SunOS 5 == Solaris 2
                osname = "solaris"
                release = "%d.%s" % (int(release[0]) - 3, release[2:])
                # We can't use "platform.architecture()[0]" because a
                # bootstrap problem. We use a dict to get an error
                # if some suspicious happens.
                bitness = {2147483647: "32bit", 9223372036854775807: "64bit"}
                machine += ".%s" % bitness[sys.maxsize]
            # fall through to standard osname-release-machine representation
        elif osname[:4] == "irix":  # could be "irix64"!
            return "%s-%s" % (osname, release)
        elif osname[:3] == "aix":
            return "%s-%s.%s" % (osname, version, release)
        elif osname[:6] == "cygwin":
            osname = "cygwin"
            import re

            rel_re = re.compile(r"[\d.]+")
            m = rel_re.match(release)
            if m:
                release = m.group()
        elif osname[:6] == "darwin":
            from . import _mosx_support

            _osx_support = _mlib_(_mosx_support)(self)
            osname, release, machine = _osx_support.get_platform_osx(self.get_config_vars(), osname, release, machine)

        return "%s-%s-%s" % (osname, release, machine)

    def get_python_version(self):
        return self._PY_VERSION_SHORT

    def _print_dict(self, title, data):
        for index, (key, value) in enumerate(sorted(data.items())):
            if index == 0:
                print("%s: " % (title))
            print('\t%s = "%s"' % (key, value))

    def _main(self):
        """Display all information sysconfig detains."""
        # if '--generate-posix-vars' in sys.argv:
        #    self._generate_posix_vars()
        #    return
        print('Platform: "%s"' % self.get_platform())
        print('Python version: "%s"' % self.get_python_version())
        print('Current installation scheme: "%s"' % self._get_default_scheme())
        print()
        self._print_dict("Paths", self.get_paths())
        print()
        self._print_dict("Variables", self.get_config_vars())



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from myrrh.framework.mpython import mbuiltins

__mlib__ = "AbcSupport"


class _interface(ABC):
    import test.support as local_support

    catch_unraisable_exception = local_support.catch_unraisable_exception

    @abstractmethod
    def findfile(self, filename, subdir=None) -> str:
        ...

    @abstractmethod
    def requires(self, resource, msg=None) -> None:
        ...

    @abstractmethod
    def gc_collect(self) -> None:
        ...

    @abstractmethod
    def detect_api_mismatch(self, ref_api, other_api, *, ignore=()) -> set[str]:
        ...

    @abstractmethod
    def cpython_only(self, test) -> typing.Any:
        ...

    @abstractmethod
    def swap_attr(self, obj, attr, new_val) -> typing.ContextManager[typing.Any | None]:
        ...

    @abstractmethod
    def no_tracing(self, func) -> typing.Callable:
        ...

    @abstractmethod
    def is_resource_enabled(self, resource) -> bool:
        ...

    @abstractmethod
    def requires_resource(self, resource) -> typing.Any:
        ...

    @abstractmethod
    def requires_fork(self) -> typing.Any:
        ...

    @abstractmethod
    def bigmemtest(self, size, memuse, dry_run=True) -> typing.Callable:
        ...

    @abstractmethod
    def requires_gzip(self, size, memuse, dry_run=True) -> typing.Callable:
        ...

    @abstractmethod
    def requires_zlib(self, size, memuse, dry_run=True) -> typing.Callable:
        ...

    @abstractmethod
    def requires_bz2(self, size, memuse, dry_run=True) -> typing.Callable:
        ...

    @abstractmethod
    def requires_lzma(self, size, memuse, dry_run=True) -> typing.Callable:
        ...

    @abstractmethod
    def requires_subprocess(self) -> typing.Callable:
        ...

    @abstractmethod
    def _force_run(self, path, func, *args) -> typing.Callable:
        ...

    @abstractmethod
    def skip_if_pgo_task(self) -> typing.Callable:
        ...

    @abstractmethod
    def refcount_test(self) -> typing.Callable:
        ...

    @abstractmethod
    def reap_children(self) -> typing.Callable:
        ...

    @abstractmethod
    def get_attribute(self, obj, name) -> typing.Any:
        ...

    @abstractmethod
    def captured_stdout(self) -> typing.Any:
        ...

    @abstractmethod
    def check__all__(self, test_case, module, name_of_module=None, extra=(), not_exported=()) -> typing.Any:
        ...

    @abstractmethod
    def check_impl_detail(self, **guards) -> bool:
        ...

    @abstractmethod
    def check_sanitizer(self, *, address=False, memory=False, ub=False) -> bool:
        ...

    @abstractmethod
    def requires_mac_ver(self, *min_version) -> typing.Callable:
        ...

    @abstractmethod
    def captured_stderr(self) -> typing.Any:
        ...

    @property
    @abstractmethod
    def PGO(self) -> bool:
        ...

    @property
    @abstractmethod
    def has_subprocess_support(self) -> bool:
        ...

    @property
    @abstractmethod
    def _4G(self):
        return self.local_support._4G

    @property
    @abstractmethod
    def PIPE_MAX_SIZE(self):
        return self.local_support.PIPE_MAX_SIZE

    @property
    @abstractmethod
    def use_resources(self) -> list | None:
        ...

    @property
    @abstractmethod
    def real_max_memuse(self) -> int:
        ...

    @property
    @abstractmethod
    def verbose(self) -> local_support.verbose:
        ...

    @property
    @abstractmethod
    def is_emscripten(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_android(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_wasi(self) -> bool:
        ...


class AbcSupport(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtest.msupport"

    __all__ = [
        # globals
        "PIPE_MAX_SIZE",
        "verbose",
        "max_memuse",
        "use_resources",
        "failfast",
        # exceptions
        "Error",
        "TestFailed",
        "TestDidNotRun",
        "ResourceDenied",
        # io
        "record_original_stdout",
        "get_original_stdout",
        "captured_stdout",
        "captured_stdin",
        "captured_stderr",
        # unittest
        "is_resource_enabled",
        "requires",
        "requires_freebsd_version",
        "requires_linux_version",
        "requires_mac_ver",
        "check_syntax_error",
        "BasicTestRunner",
        "run_unittest",
        "run_doctest",
        "requires_gzip",
        "requires_bz2",
        "requires_lzma",
        "bigmemtest",
        "bigaddrspacetest",
        "cpython_only",
        "get_attribute",
        "requires_IEEE_754",
        "requires_zlib",
        "has_fork_support",
        "requires_fork",
        "has_subprocess_support",
        "requires_subprocess",
        "has_socket_support",
        "requires_working_socket",
        "anticipate_failure",
        "load_package_tests",
        "detect_api_mismatch",
        "check__all__",
        "skip_if_buggy_ucrt_strfptime",
        "check_disallow_instantiation",
        "check_sanitizer",
        "skip_if_sanitizer",
        # sys
        "is_jython",
        "is_android",
        "is_emscripten",
        "is_wasi",
        "check_impl_detail",
        "unix_shell",
        "setswitchinterval",
        # network
        "open_urlresource",
        # processes
        "reap_children",
        # miscellaneous
        "run_with_locale",
        "swap_item",
        "findfile",
        "infinite_recursion",
        "swap_attr",
        "Matcher",
        "set_memlimit",
        "SuppressCrashReport",
        "sortdict",
        "run_with_tz",
        "PGO",
        "missing_compiler_executable",
        "ALWAYS_EQ",
        "NEVER_EQ",
        "LARGEST",
        "SMALLEST",
        "LOOPBACK_TIMEOUT",
        "INTERNET_TIMEOUT",
        "SHORT_TIMEOUT",
        "LONG_TIMEOUT",
    ]

    __doc__ = _interface.local_support.__doc__

    __delegated__ = {_interface: _interface.local_support}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_support, self)

        self.__delegate__(_interface, mod)

    def local_findfile(self, filename, subdir=None):
        import os
        import sys
        import test.support

        TEST_SUPPORT_DIR = os.path.dirname(os.path.abspath(test.support.__file__))
        TEST_HOME_DIR = os.path.dirname(TEST_SUPPORT_DIR)

        if os.path.isabs(filename):
            return filename

        if subdir is not None:
            filename = os.path.join(subdir, filename)
        path = [TEST_HOME_DIR] + sys.path
        for dn in path:
            fn = os.path.join(dn, filename)
            if os.path.exists(fn):
                return fn

        return filename

    def findfile(self, filename, subdir=None):
        _cast_ = self.myrrh_os.fscast(filename)
        local_filename = self.local_findfile(filename, subdir)
        try:
            with open(local_filename) as f:
                self.myrrh_syscall.stream_out(self.myrrh_os.fsencode(filename), f.buffer)
        except FileNotFoundError:
            pass

        return _cast_(self.myrrh_os.getpathb(filename))



import os
import errno as _errno

__all__ = (
    "MOsError",
    "MIOException",
    "UnexpectedException",
    "FileException",
    "ExecutionFailure",
    "ExecutionFailureCauseErr",
    "ExecutionFailureCauseRVal",
)


class MOsError(OSError):
    def __init__(self, system, errno=-1, strerror="", exc=None, error_translate=None, args=()):
        if strerror and error_translate is None and system:
            if hasattr(system, "error_translate"):
                error_translate = system.error_translate

        if exc is not None:
            errno = exc.errno if hasattr(exc, "errno") and exc.errno is not None else errno
            strerror = exc.strerror if hasattr(exc, "strerror") and exc.strerror is not None else "(%s)" % exc.__class__ + str(exc)
        elif strerror and errno == -1 and error_translate is not None:
            errno = error_translate(strerror)

        super().__init__(errno, strerror, *args)

        if errno is not None and strerror:
            self.strerror = os.strerror(errno) if errno >= 0 else ("undefined exception with %s" % (self.args,))

        self.system = system
        self.exc = exc
        self.args = args

    def raised(self):
        exc_class = OSError(self.errno, self.strerror).__class__
        if not issubclass(MOsError, exc_class):
            exc = type(exc_class.__name__, (exc_class, MOsError), dict())(self.errno, self.strerror, *self.args)
        else:
            exc = self

        __traceback__ = self.exc.__traceback__ if self.exc else None

        raise exc.with_traceback(__traceback__)


class MIOException(MOsError, IOError):
    def __init__(
        self,
        system,
        errno=-1,
        strerror="",
        exc=None,
        filename=None,
        error_translate=None,
    ):
        MOsError.__init__(
            self,
            system,
            errno,
            strerror,
            exc,
            args=(filename,),
            error_translate=error_translate,
        )
        if errno is not None and len(strerror) == 0:
            self.strerror = os.strerror(errno) if errno >= 0 else "undefined exception"

        self.system = system


class UnexpectedException(MOsError):
    def __init__(self, system, errno=0, strerror="unexpected exception"):
        super().__init__(errno, strerror)


class FileException(MOsError):
    def __init__(self, system, errno=0, filename=""):
        strerror = os.strerror(errno or _errno.ENOENT)

        if filename:
            filename = system.myrrh_os.fsdecode(filename)
            args = (str(filename),)

        super().__init__(system, errno=errno or _errno.ENOENT, strerror=strerror, args=args)


class ExecutionFailure(MOsError):
    def __init__(self, system, val, expected_val, msg, *args, errno=-1, error_translate=None):
        super().__init__(system, errno, strerror=msg, error_translate=error_translate, args=args)
        self.__val = val
        try:
            iter(expected_val)
            self.__expected_val = expected_val
        except TypeError:
            self.__expected_val = (expected_val,)

    def check(self):
        if self.__val not in self.__expected_val:
            self.raised()


class ExecutionFailureCauseErr(ExecutionFailure):
    def __init__(self, system, err, expected_err, *args, errno=-1):
        super().__init__(system, err, expected_err, err, *args, errno=errno)
        self.err = err
        self.expected_err = expected_err


class ExecutionFailureCauseRVal(ExecutionFailure):
    def __init__(self, system, err, rval, expected_rval, *args, errno=-1, error_translate=None):
        super().__init__(system, rval, expected_rval, err, *args, errno=errno, error_translate=error_translate)

        self.err = err
        self.rval = rval
        self.expected_rval = expected_rval
